import asyncio
import time
import unittest
import unittest.mock

import mido  # type: ignore
import shc.interfaces.midi
from shc.datatypes import RangeUInt8

from .._helper import InterfaceThreadRunner, AsyncMock


try:
    mido.backend.load()
    mido_backend_available = True
    mido_backend_error = ""
except Exception as e:
    mido_backend_available = False
    mido_backend_error = str(e)


class MIDITest(unittest.TestCase):
    def test_errors(self) -> None:
        with self.assertRaises(ValueError):
            interface = shc.interfaces.midi.MidiInterface()

        interface = shc.interfaces.midi.MidiInterface("foo")
        var1 = interface.note_on_off(42)
        var2 = interface.note_on_off(42)
        self.assertIs(var1, var2)
        with self.assertRaises(ValueError):
            interface.note_velocity(42)


@unittest.skipUnless(mido_backend_available, "mido MIDI backend is not awailable: {}".format(mido_backend_error))
class MIDIInputTest(unittest.TestCase):
    def setUp(self) -> None:
        self.port_name = 'TestOutPort' + self.id().split('.')[-1]
        self.dummy_port = mido.open_output(self.port_name, virtual=True)

        self.interface_runner = InterfaceThreadRunner(shc.interfaces.midi.MidiInterface, self.port_name, None)
        self.interface = self.interface_runner.interface

    def tearDown(self) -> None:
        self.interface_runner.stop()
        self.dummy_port.close()

    def test_input(self) -> None:
        var1 = self.interface.note_on_off(5)
        var2 = self.interface.note_velocity(7)
        var3 = self.interface.control_change(1)

        with unittest.mock.patch.object(var1, '_publish') as publish_mock1,\
             unittest.mock.patch.object(var2, '_publish') as publish_mock2,\
             unittest.mock.patch.object(var3, '_publish') as publish_mock3:

            self.interface_runner.start()
            time.sleep(0.05)

            # Things to be ignored
            self.dummy_port.send(mido.Message('note_on', channel=0, note=42, velocity=20))
            self.dummy_port.send(mido.Message('control_change', channel=0, control=42, value=0))
            self.dummy_port.send(mido.Message('aftertouch', channel=0, value=18))
            time.sleep(0.05)

            publish_mock1.assert_not_called()
            publish_mock2.assert_not_called()
            publish_mock3.assert_not_called()

            # Note on
            self.dummy_port.send(mido.Message('note_on', channel=0, note=5, velocity=20))
            self.dummy_port.send(mido.Message('note_on', channel=0, note=7, velocity=20))
            time.sleep(0.05)

            publish_mock1.assert_called_once_with(True, unittest.mock.ANY)
            publish_mock2.assert_called_once_with(RangeUInt8(40), unittest.mock.ANY)
            publish_mock3.assert_not_called()

            # Note off
            publish_mock1.reset_mock()
            publish_mock2.reset_mock()
            self.dummy_port.send(mido.Message('note_off', channel=0, note=5, velocity=40))
            self.dummy_port.send(mido.Message('note_off', channel=0, note=7, velocity=40))
            time.sleep(0.05)

            publish_mock1.assert_called_once_with(False, unittest.mock.ANY)
            publish_mock2.assert_called_once_with(RangeUInt8(80), unittest.mock.ANY)
            publish_mock3.assert_not_called()

            # Control change
            publish_mock1.reset_mock()
            publish_mock2.reset_mock()
            self.dummy_port.send(mido.Message('control_change', channel=0, control=1, value=42))
            time.sleep(0.05)

            publish_mock1.assert_not_called()
            publish_mock2.assert_not_called()
            publish_mock3.assert_called_once_with(RangeUInt8(84), unittest.mock.ANY)

    def test_emulated_toggle(self) -> None:
        var1 = self.interface.note_on_off(5, emulate_toggle=True)

        with unittest.mock.patch.object(var1, '_publish') as publish_mock:

            self.interface_runner.start()
            time.sleep(0.05)

            # Toggle on
            self.dummy_port.send(mido.Message('note_on', channel=0, note=5, velocity=127))
            time.sleep(0.05)
            publish_mock.assert_called_once_with(True, unittest.mock.ANY)

            publish_mock.reset_mock()
            self.dummy_port.send(mido.Message('note_off', channel=0, note=5, velocity=0))
            time.sleep(0.05)
            publish_mock.assert_called_once_with(True, unittest.mock.ANY)

            # TODO test button feedback

            # Toggle off
            publish_mock.reset_mock()
            self.dummy_port.send(mido.Message('note_on', channel=0, note=5, velocity=127))
            time.sleep(0.05)
            publish_mock.assert_called_once_with(False, unittest.mock.ANY)

            publish_mock.reset_mock()
            self.dummy_port.send(mido.Message('note_off', channel=0, note=5, velocity=0))
            time.sleep(0.05)
            publish_mock.assert_called_once_with(False, unittest.mock.ANY)

            asyncio.run_coroutine_threadsafe(var1.write(True, [self]), self.interface_runner.loop)
            time.sleep(0.05)

            # Toggle off again
            publish_mock.reset_mock()
            self.dummy_port.send(mido.Message('note_on', channel=0, note=5, velocity=127))
            time.sleep(0.05)
            publish_mock.assert_called_once_with(False, unittest.mock.ANY)


@unittest.skipUnless(mido_backend_available, "mido MIDI backend is not awailable: {}".format(mido_backend_error))
class MIDIOutputTest(unittest.TestCase):
    def setUp(self) -> None:
        self.port_name = 'TestInPort' + self.id().split('.')[-1]
        self.callback = unittest.mock.Mock()
        self.dummy_port = mido.open_input(self.port_name, virtual=True, callback=self.callback)

        self.interface_runner = InterfaceThreadRunner(shc.interfaces.midi.MidiInterface, None, self.port_name,
                                                      send_channel=9)
        self.interface = self.interface_runner.interface

    def tearDown(self) -> None:
        self.interface_runner.stop()
        self.dummy_port.close()

    def test_output(self) -> None:
        var1 = self.interface.note_on_off(5)
        var2 = self.interface.note_velocity(7)
        var3 = self.interface.control_change(1)

        self.interface_runner.start()
        time.sleep(0.05)

        # Note on
        asyncio.run_coroutine_threadsafe(var1.write(True, [self]), self.interface_runner.loop)
        time.sleep(0.05)
        self.callback.assert_called_once()
        message = self.callback.call_args[0][0]
        assert isinstance(message, mido.Message)
        self.assertEqual(message.channel, 9)
        self.assertEqual(message.type, "note_on")
        self.assertEqual(message.note, 5)
        self.assertNotEqual(message.velocity, 0)

        # Note off
        self.callback.reset_mock()
        asyncio.run_coroutine_threadsafe(var1.write(False, [self]), self.interface_runner.loop)
        time.sleep(0.05)
        self.callback.assert_called_once()
        message = self.callback.call_args[0][0]
        assert isinstance(message, mido.Message)
        self.assertEqual(message.channel, 9)
        self.assertEqual(message.type, "note_off")
        self.assertEqual(message.note, 5)
        self.assertEqual(message.velocity, 0)

        # Note on (velocity)
        self.callback.reset_mock()
        asyncio.run_coroutine_threadsafe(var2.write(RangeUInt8(42), [self]), self.interface_runner.loop)
        time.sleep(0.05)
        self.callback.assert_called_once()
        message = self.callback.call_args[0][0]
        assert isinstance(message, mido.Message)
        self.assertEqual(message.channel, 9)
        self.assertEqual(message.type, "note_on")
        self.assertEqual(message.note, 7)
        self.assertEqual(message.velocity, 21)

        # Note off (velocity)
        self.callback.reset_mock()
        asyncio.run_coroutine_threadsafe(var2.write(RangeUInt8(0), [self]), self.interface_runner.loop)
        time.sleep(0.05)
        self.callback.assert_called_once()
        message = self.callback.call_args[0][0]
        assert isinstance(message, mido.Message)
        self.assertEqual(message.channel, 9)
        self.assertEqual(message.type, "note_off")
        self.assertEqual(message.note, 7)
        self.assertEqual(message.velocity, 0)

        # Control change
        self.callback.reset_mock()
        asyncio.run_coroutine_threadsafe(var3.write(RangeUInt8(56), [self]), self.interface_runner.loop)
        time.sleep(0.05)
        self.callback.assert_called_once()
        message = self.callback.call_args[0][0]
        assert isinstance(message, mido.Message)
        self.assertEqual(message.channel, 9)
        self.assertEqual(message.type, "control_change")
        self.assertEqual(message.control, 1)
        self.assertEqual(message.value, 28)



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import asyncio
import json
import logging
import weakref
from json import JSONDecodeError
from typing import Type, Dict, Generic, List, Any, Optional

import aiohttp

from ._helper import SupervisedClientInterface
from ..base import T, Subscribable, Writable, Readable, UninitializedError
from ..conversion import SHCJsonEncoder, from_json
from ..supervisor import register_interface, stop

logger = logging.getLogger(__name__)


# Timeout for read, write and subscribe calls in seconds
TIMEOUT = 5.0


class SHCWebClient(SupervisedClientInterface):
    """
    Client for connecting to remote SHC instances via the websocket API, provided by :class:`shc.web.WebServer`

    For each API object (:class:`shc.web.WebApiObject`) on the remote server, a local "proxy" object can be created by
    calling :meth:`object` with the server object's name and correct type. This object (:class:`WebApiClientObject`)
    forwards local `read` and `write` calls to the server and remote writes from the server to locally subscribed
    objects.

    :param server: Base URL of the remote SHC webserver instance without trailing slash, e.g. 'https://example.com/shc'.
        The path of the API websocket ('/api/v1/ws') is appended internally.
    :param auto_reconnect: If True (default), the API client tries to reconnect automatically on connection errors with
        exponential backoff (1 * 1.25^n seconds sleep). Otherwise, the complete SHC system is shut down, when a
        connection error occurs.
    :param failsafe_start: If True and auto_reconnect is True, the API client allows SHC to start up, even if the API
        connection can not be established in the first try. The connection is retried in background with exponential
        backoff (see `auto_reconnect` option). Otherwise (default), the first connection attempt on startup is not
        retried and will shutdown the SHC application on failure, even if `auto_reconnect` is True.
    """
    def __init__(self, server: str, auto_reconnect: bool = True, failsafe_start: bool = False) -> None:
        super().__init__(auto_reconnect, failsafe_start)
        self.server = server
        self._api_objects: Dict[str, WebApiClientObject] = {}

        self._session: aiohttp.ClientSession
        self._ws: Optional[aiohttp.ClientWebSocketResponse] = None
        self._waiting_futures: weakref.WeakValueDictionary[int, asyncio.Future] = weakref.WeakValueDictionary()

    async def _subscribe_and_wait(self, name: str) -> None:
        """
        Internal coroutine for subscribing for a specified API object on the remote SHC server.

        This method sends the 'subscribe' message to the server and awaits the response (via an asyncio Future). To
        receive the response, it must only be used *after* starting the :meth:`run` coroutine in a parallel task. The
        coroutine raises an exception when a negative response is reiceved from the server or no response is received at
        all within TIMEOUT seconds.

        :param name: Name (designator) of the server's API object to be subscribed
        :raises WebSocketAPIError: when the server responds with a non-200 status code
        :raises asyncio.TimeoutError: when no response is received within TIMEOUT seconds
        """
        future = asyncio.get_event_loop().create_future()
        self._waiting_futures[id(future)] = future
        assert self._ws is not None
        await self._ws.send_json({'action': 'subscribe', 'name': name, 'handle': id(future)})
        result = await asyncio.wait_for(future, TIMEOUT)
        if not 200 <= result['status'] < 300:
            raise WebSocketAPIError("Failed to subscribe SHC API object '{}' with status {}: {}"
                                    .format(name, result['status'], result.get('error')))

    async def start(self) -> None:
        self._session = aiohttp.ClientSession()
        await super().start()

    async def _connect(self) -> None:
        # TODO change timeout: https://docs.aiohttp.org/en/stable/client_quickstart.html#timeouts
        self._ws = await self._session.ws_connect(self.server + '/api/v1/ws')

    async def _subscribe(self) -> None:
        await asyncio.gather(*(self._subscribe_and_wait(name)
                               for name, obj in self._api_objects.items()
                               if obj._subscribers or obj._triggers))
        # TODO gather results and give a better error description

    async def _disconnect(self) -> None:
        logger.info("Closing client websocket to %s ...", self.server)
        if self._ws is not None:
            await self._ws.close()

    async def stop(self) -> None:
        await super().stop()
        await self._session.close()

    async def _run(self) -> None:
        """
        Entrypoint for the async "run" Task for receiving Websocket messages
        """
        assert self._ws is not None
        self._running.set()

        # Receive websocket messages until websocket is closed
        msg: aiohttp.WSMessage
        async for msg in self._ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                self._websocket_dispatch(msg)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                logger.error('SHC API websocket failed with %s', self._ws.exception())

        logger.debug('SHC API websocket connection closed')

    def _websocket_dispatch(self, msg: aiohttp.WSMessage) -> None:
        try:
            message = msg.json()
        except JSONDecodeError:
            logger.warning("Websocket message from SHC server is not a valid JSON string: %s", msg.data)
            return

        logger.debug("Incoming message from websocket API: %s", message)

        try:
            name = message["name"]
            status = message["status"]
        except KeyError:
            logger.warning("Websocket message from SHC server does not include 'name' and 'status' fields: %s",
                           msg.data)
            return

        # If the message has a handle and the handle is associated with a waiting future, set the message as result
        if 'handle' in message:
            future = self._waiting_futures.get(message['handle'])
            if future is None:
                logger.info("Received websocket API message with handle, which refers to non-existent future: %s", msg)
            else:
                future.set_result(message)
                del self._waiting_futures[id(future)]

        # New (or initial) value for subscribed object (not on read response)
        if 'value' in message and ('action' not in message or message['action'] == 'subscribe'):
            # We don't need to do this in an asynchronous task, since the new_value method uses asynchronous publishing
            self._api_objects[name].new_value(message['value'])

        elif 'handle' not in message:
            logger.warning("Received unexpected message from SHC websocket API: %s", msg)

    async def _send_value(self, name: str, value: Any) -> None:
        """
        Coroutine called by WebApiClientObject's _write() method to send a new value to the remote SHC server

        The method awaits the receipt of the server's answer or a timeout of TIMEOUT seconds. In case of a server side
        error or a response timeout, an exception is raised.

        :param name: Name of the API object
        :param value: The value to be sent to the server. The value's type must match the server-side API object's type
            and be encodable with the SHCJsonEncoder (i.e. have a json conversion registered in the
            :mod:`shc.conversion` module).
        :raises WebSocketAPIError: when sending the new value fails on the server side (i.e. the API's return code is
            not in the 200-range). This may be caused by a type or object name mismatch.
        :raises asyncio.TimeoutError: when no response is received from the server within TIMEOUT seconds
        """
        if self._ws is None:
            raise RuntimeError("Websocket of SHC client for API at {} has not been connected yet".format(self.server))
        future = asyncio.get_event_loop().create_future()
        self._waiting_futures[id(future)] = future
        logger.debug("Writing value from SHC API object %s ...", name)
        await self._ws.send_str(json.dumps({'action': 'post', 'name': name, 'value': value, 'handle': id(future)},
                                           cls=SHCJsonEncoder))

        result = await asyncio.wait_for(future, TIMEOUT)
        if 200 <= result['status'] < 300:
            logger.debug("Writing value to SHC API object %s succeeded", name)
        else:
            raise WebSocketAPIError("Writing value to SHC API failed with error {}: {}"
                                    .format(result['status'], result.get('error')))

    async def _read_value(self, name: str) -> Any:
        """
        Coroutine called by WebApiClientObject's read() to fetch the current value of an API object from the remote SHC
        server

        The method awaits the server's answer and returns the raw json-decoded value. If no response is received after
        TIMEOUT seconds the coroutine aborts with an exception.

        :param name: Name of the API object to retrieve its value
        :returns: The json-decoded, but not yet converted value of the API object. It should be converted to the
            expected type, using :func:`shc.conversion.from_json`.
        :raises UninitializedError: when a status code 409 is received from the server, indicating that the API object
            (resp. the underlying readable object) has not yet been initialized.
        :raises WebSocketAPIError: when reading the current value fails on the server side (i.e. the API's return code
            is not in the 200-range). This may be caused object name mismatch or another server-side error.
        :raises asyncio.TimeoutError: when no response is received from the server within TIMEOUT seconds
        """
        if self._ws is None:
            raise RuntimeError("Websocket of SHC client for API at {} has not been connected yet".format(self.server))
        future = asyncio.get_event_loop().create_future()
        self._waiting_futures[id(future)] = future
        logger.debug("Reading value from SHC API object %s ...", name)
        await self._ws.send_str(json.dumps({'action': 'get', 'name': name, 'handle': id(future)}, cls=SHCJsonEncoder))
        result = await asyncio.wait_for(future, TIMEOUT)
        if 200 <= result['status'] < 300:
            logger.debug("Read value %s from SHC API for object %s", result['value'], name)
            return result['value']
        elif result['status'] == 409:
            logger.debug("'get' action of SHC API for object %s returned uninitialized (status 409)", name)
            raise UninitializedError
        else:
            raise WebSocketAPIError("Writing value to SHC API failed with error {}: {}"
                                    .format(result['status'], result.get('error')))

    def object(self, type_: Type, name: str) -> "WebApiClientObject":
        """
        Create a `Connectable` object for communicating with an API object on the remote SHC server.

        The returned object is `Subscribable`, `Readable` and `Writable`. Read and write calls are basically forwarded
        to the corresponding :class:`shc.web.WebApiObject`; new values, pushed from the server, are published to the
        object's subscribers.

        :param type_: The data type of the server's API object. This is also the `type` of the returned `Connectable`
            object.
        :param name: The name of the server's API object
        """
        if name in self._api_objects:
            existing = self._api_objects[name]
            if existing.type is not type_:
                raise TypeError("Type {} does not match type {} of existing API client object with same name"
                                .format(type_, existing.type))
            return existing
        else:
            api_object = WebApiClientObject(self, type_, name)
            self._api_objects[name] = api_object
            return api_object

    def __repr__(self) -> str:
        return "{}(server={})".format(self.__class__.__name__, self.server)


class WebApiClientObject(Readable[T], Writable[T], Subscribable[T], Generic[T]):
    """
    A `Connectable` object to communicate with a single API object of a remote SHC websocket API.

    This object basically builds a transparent tunnel to the corresponding :class:`shc.web.WebApiObject` with the
    matching `name` on the server side, forwarding :meth:`read` and :meth:`write` calls to connected objects on the
    server and publishing received value updates (from `write` calls on the server) to locally connected objects.

    Thus, this class inherits from :class:`shc.base.Readable`, :class:`shc.base.Wrtiable` and
    :class:`shc.base.Subscribable`.
    """
    _stateful_publishing = True

    def __init__(self, client: SHCWebClient, type_: Type[T], name: str):
        self.type = type_
        super().__init__()
        self.client = client
        self.name = name

    async def read(self) -> T:
        return from_json(self.type, await self.client._read_value(self.name))

    async def _write(self, value: T, origin: List[Any]) -> None:
        # Asynchronous local feedback publishing. This ensures that conflicting updates, which are currently waiting for
        # local processing by a subscriber can be corrected by resetting this update's origin.
        self._publish(value, origin)
        await self.client._send_value(self.name, value)

    def new_value(self, value: Any) -> None:
        """
        Called by the associated :class:`SHCWebClient` when a new value for this object is received from the remote SHC
        server to be published to local subscribers.

        :param value: The received, json-decoded value. It will be processed using :meth:`shc.conversion.from_json`.
        """
        try:
            self._publish(from_json(self.type, value), [])
            logger.debug("Received new value %s for SHC API object %s", value, self.name)
        except Exception as e:
            logger.error("Error while processing new value %s for API object %s", value, self.name, exc_info=e)


class WebSocketAPIError(RuntimeError):
    """
    Exception to be raised by :meth:`WebApiClientObject.read`, :meth:`WebApiClientObject.write` and on startup of the
    :class:`SHCWebClient`, when an SHC websocket API action fails with an non-200 status code.
    """
    pass



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import enum
import json
import datetime
import logging
import math
from typing import Type, Generic, List, Any, Optional, Set, Tuple, Union, cast, TypeVar, Callable

import aiohttp.web

from .. import timer
from ..base import T, Readable, Writable, UninitializedError
from ..conversion import SHCJsonEncoder
from ..web.interface import WebUIConnector

logger = logging.getLogger(__name__)


class AggregationMethod(enum.Enum):
    AVERAGE = 0
    MINIMUM = 1
    MAXIMUM = 2
    ON_TIME = 3
    ON_TIME_RATIO = 4


class PersistenceVariable(Readable[T], Writable[T], Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, type_: Type[T], log: bool):
        self.type = type_
        super().__init__()
        self.log = log
        self.subscribed_web_ui_views: List[LoggingRawWebUIView] = []

    @abc.abstractmethod
    async def _read_from_log(self) -> Optional[T]:
        pass

    @abc.abstractmethod
    async def _write_to_log(self, value: T) -> None:
        pass

    @abc.abstractmethod
    async def retrieve_log(self, start_time: datetime.datetime, end_time: datetime.datetime,
                           include_previous: bool = True) -> List[Tuple[datetime.datetime, T]]:
        """
        Retrieve all log entries for this log variable in the specified time range from the log backend/database

        The method shall return a list of all log entries with a timestamp greater or equal to the `start_time` and
        less than the `end_time`. If `include_previous` is True (shall be the default value), the last entry *before*
        the start shall also be included, if there is no entry exactly at the start_time.

        :param start_time: Begin of the time range (inclusive)
        :param end_time: End of the time range (exclusive)
        :param include_previous: If True (the default), the last value *before* `start_time`
        """
        pass

    async def retrieve_aggregated_log(self, start_time: datetime.datetime, end_time: datetime.datetime,
                                      aggregation_method: AggregationMethod, aggregation_interval: datetime.timedelta
                                      ) -> List[Tuple[datetime.datetime, float]]:
        data = await self.retrieve_log(start_time, end_time, include_previous=True)
        aggregation_timestamps = [start_time + i * aggregation_interval
                                  for i in range(math.ceil((end_time - start_time) / aggregation_interval))]

        # The last aggregation_timestamps is not added to the results, but only used to delimit the last aggregation
        # interval.
        if aggregation_timestamps[-1] < end_time:
            aggregation_timestamps.append(end_time)

        if aggregation_method in (AggregationMethod.MINIMUM, AggregationMethod.MAXIMUM, AggregationMethod.AVERAGE):
            if not issubclass(self.type, (int, float)):
                raise TypeError("min/max/avg. aggregation is only applicable to int and float type log variables.")

        elif aggregation_method not in (AggregationMethod.ON_TIME, AggregationMethod.ON_TIME_RATIO):
            raise ValueError("Unsupported aggregation method {}".format(aggregation_method))

        aggregator_t = {
            AggregationMethod.MINIMUM: MinAggregator,
            AggregationMethod.MAXIMUM: MaxAggregator,
            AggregationMethod.AVERAGE: AverageAggregator,
            AggregationMethod.ON_TIME: OnTimeAggregator,
            AggregationMethod.ON_TIME_RATIO: OnTimeRatioAggregator,
        }[aggregation_method]()  # type: ignore
        aggregator = cast(AbstractAggregator[T, Any], aggregator_t)  # This is guaranteed by our type checks above

        next_aggr_ts_index = 0
        # Get first entry and its timestamp for skipping of empty aggregation intervals and initialization of the
        # first relevant interval
        iterator = iter(data)
        try:
            last_ts, last_value = next(iterator)
        except StopIteration:
            return []

        # Ignore aggregation intervals before the first entry
        while last_ts >= aggregation_timestamps[next_aggr_ts_index]:
            next_aggr_ts_index += 1
            if next_aggr_ts_index >= len(aggregation_timestamps):
                return []

        result: List[Tuple[datetime.datetime, float]] = []
        aggregator.reset()

        for ts, value in iterator:
            # The timestamp is after the next aggregation timestamp, finalize the current aggregation timestamp
            # value
            if ts >= aggregation_timestamps[next_aggr_ts_index]:
                if next_aggr_ts_index > 0:
                    # Add remaining part to the last aggregation interval
                    aggregator.aggregate(last_ts, aggregation_timestamps[next_aggr_ts_index], last_value)

                    # Add average result entry from accumulated values
                    result.append((aggregation_timestamps[next_aggr_ts_index-1], aggregator.get()))

                next_aggr_ts_index += 1
                if next_aggr_ts_index >= len(aggregation_timestamps):
                    return result

                # Do the same for every following empty interval
                while ts >= aggregation_timestamps[next_aggr_ts_index]:
                    aggregator.reset()
                    aggregator.aggregate(aggregation_timestamps[next_aggr_ts_index-1],
                                         aggregation_timestamps[next_aggr_ts_index], last_value)
                    result.append((aggregation_timestamps[next_aggr_ts_index-1], aggregator.get()))

                    next_aggr_ts_index += 1
                    if next_aggr_ts_index >= len(aggregation_timestamps):
                        return result

                aggregator.reset()

            # Accumulate the weighted value and time interval to the `*_sum` variables
            if next_aggr_ts_index > 0:
                interval_start = aggregation_timestamps[next_aggr_ts_index-1]
                aggregator.aggregate(max(last_ts, interval_start), ts, last_value)

            last_value = value
            last_ts = ts

        if next_aggr_ts_index > 0:
            # Add remaining part to the last aggregation interval
            aggregator.aggregate(last_ts, aggregation_timestamps[next_aggr_ts_index], last_value)

            # Add average result entry from accumulated values
            result.append((aggregation_timestamps[next_aggr_ts_index-1], aggregator.get()))

        next_aggr_ts_index += 1

        # Fill remaining aggregation intervals
        while next_aggr_ts_index < len(aggregation_timestamps):
            aggregator.reset()
            aggregator.aggregate(aggregation_timestamps[next_aggr_ts_index - 1],
                                 aggregation_timestamps[next_aggr_ts_index], last_value)
            result.append((aggregation_timestamps[next_aggr_ts_index - 1], aggregator.get()))

            next_aggr_ts_index += 1
        return result

    async def read(self) -> T:
        value = await self._read_from_log()
        if value is None:
            raise UninitializedError("No value for has been persisted for variable '{}' yet.".format(self))
        return value

    async def _write(self, value: T, origin: List[Any]):
        logger.debug("%s value %s for %s to log backend", "logging" if self.log else "updating", value, self)
        await self._write_to_log(value)
        for web_ui_view in self.subscribed_web_ui_views:
            await web_ui_view.new_value(datetime.datetime.now(datetime.timezone.utc), value)


class LoggingRawWebUIView(WebUIConnector):
    """
    A WebUIConnector which is used to retrieve a log/timeseries of a certain log variable for a certain time
    interval via the Webinterface UI websocket and subscribe to updates of that log variable.
    """
    def __init__(self, variable: PersistenceVariable, interval: datetime.timedelta,
                 converter: Optional[Callable] = None, include_previous: bool = True):
        if not variable.log:
            raise ValueError("Cannot use a PersistenceVariable with log=False for a web logging web ui widget")
        super().__init__()
        self.variable = variable
        variable.subscribed_web_ui_views.append(self)
        self.interval = interval
        self.subscribed_websockets: Set[aiohttp.web.WebSocketResponse] = set()
        self.converter: Callable = converter or (lambda x: x)
        self.include_previous = include_previous

    async def new_value(self, timestamp: datetime.datetime, value: Any) -> None:
        """
        Coroutine to be called by the class:`PersistenceVariable` this view belongs to, when it receives and logs a new
        value, so the value can instantly be plotted by all subscribed web clients.

        :param timestamp: Exact timestamp of the new value
        :param value: The new value
        """
        await self._websocket_publish({
            'init': False,
            'data': [(timestamp, self.converter(value))]
        })

    async def _websocket_before_subscribe(self, ws: aiohttp.web.WebSocketResponse) -> None:
        data = await self.variable.retrieve_log(datetime.datetime.now(datetime.timezone.utc)
                                                - self.interval,
                                                datetime.datetime.now(datetime.timezone.utc)
                                                + datetime.timedelta(seconds=5), include_previous=self.include_previous)
        await ws.send_str(json.dumps({'id': id(self),
                                      'v': {
                                          'init': True,
                                          'data': [(ts, self.converter(v)) for ts, v in data]
                                      }},
                                     cls=SHCJsonEncoder))


class LoggingAggregatedWebUIView(WebUIConnector):
    """
    A WebUIConnector which is used to retrieve and periodically update aggregated log data of a certain log variable for
    a certain time interval via the Webinterface UI websocket.
    """
    def __init__(self, variable: PersistenceVariable, interval: datetime.timedelta, aggregation: AggregationMethod,
                 aggregation_interval: datetime.timedelta, converter: Optional[Callable] = None,
                 align_to: datetime.datetime = datetime.datetime(2020, 1, 1, 0, 0, 0),
                 update_interval: Optional[datetime.timedelta] = None):
        # TODO add pre_converter (e.g. for scaling or getting a single value out of Namespaces)
        # TODO add PersistenceVariable type check?
        if not variable.log:
            raise ValueError("Cannot use a PersistenceVariable with log=False for a web logging web ui widget")
        super().__init__()
        self.variable = variable
        self.interval = interval
        self.aggregation = aggregation
        self.aggregation_interval = aggregation_interval
        self.subscribed_websockets: Set[aiohttp.web.WebSocketResponse] = set()
        self.converter = converter or (lambda x: x)
        self.align_to = align_to
        update_interval = (update_interval if update_interval is not None
                           else min(aggregation_interval / 2, datetime.timedelta(minutes=1)))
        self.last_update = datetime.datetime.fromtimestamp(0).astimezone() + 2 * aggregation_interval

        self.timer = timer.Every(update_interval)
        self.timer.trigger(self._update, synchronous=True)

    async def _websocket_before_subscribe(self, ws: aiohttp.web.WebSocketResponse) -> None:
        begin, end = self.calculate_interval(False)
        data = await self.variable.retrieve_aggregated_log(begin, end, self.aggregation, self.aggregation_interval)
        # TODO use caching?
        await ws.send_str(json.dumps({'id': id(self),
                                      'v': {
                                          'init': True,
                                          'data': [(ts, self.converter(v)) for ts, v in data]
                                      }},
                                     cls=SHCJsonEncoder))

    def calculate_interval(self, only_latest: bool) -> Tuple[datetime.datetime, datetime.datetime]:
        """
        Calculate the begin and end timestamps to be passed to :meth:`PersistenceVariable.retrieve_aggregated_log` based
        on the configured interval, aggregation_interval, align timestamp.

        :param only_latest: If True, the begin timestamp is chosen in such a way, that only the new and (probably)
            changed aggregated values since the last update (see self.last_update) are returned.
        :return: (start_time, end_time) for `retrieve_aggregated_log()`
        """
        end = datetime.datetime.now().astimezone()
        preliminary_begin = end - self.interval
        if only_latest:
            preliminary_begin = max(preliminary_begin, self.last_update - self.aggregation_interval)
        align = self.align_to.astimezone()
        align_count = (preliminary_begin - align) // self.aggregation_interval
        begin = align + (align_count + 1) * self.aggregation_interval

        return begin, end

    async def _update(self, _v, _o) -> None:
        if not self.subscribed_websockets:
            return

        begin, end = self.calculate_interval(True)
        data = await self.variable.retrieve_aggregated_log(begin, end, self.aggregation, self.aggregation_interval)
        await self._websocket_publish({'init': False,
                                       'data': [(ts, self.converter(v)) for ts, v in data]})
        self.last_update = end


S = TypeVar('S')


class AbstractAggregator(Generic[T, S], metaclass=abc.ABCMeta):
    """
    Abstract base class for Aggregator classes, a special helper class, used to define the actual aggregation logic in
    :meth:`PersistenceVariable.retrieve_aggregated_log`. For each aggregation method (see :class:`AggregationMethods`),
    a specialized Aggregator class is available.

    The aggregator defines, how individual values, which lasted for a given time interval, are aggregated into a single
    value. It is controlled by the `retrieve_aggregated_log()` method via three interface methods:
    * :meth:`reset` is called at the begin of every aggregation interval to reset the aggregated value
    * :meth:`aggregate` is called for each value change with the value and the begin and end timestamps of the interval
        for which the value lasted. These time intervals can be used to calculate the value's weight in a weighted
        average or the on time of a bool aggregation
    * :meth:`get` is called at the end of every aggregation interval to retrieve the aggregated value

    """
    __slots__ = ()

    def reset(self) -> None:
        pass

    @abc.abstractmethod
    def get(self) -> S:
        pass

    @abc.abstractmethod
    def aggregate(self, start: datetime.datetime, end: datetime.datetime, value: T) -> None:
        pass


class AverageAggregator(AbstractAggregator[Union[float, int], float]):
    """
    Aggregator implementation for AggregationMethod.AVERAGE. It calculates an average of the values, weighted by their
    respective time intervals.
    """
    __slots__ = ('value_sum', 'time_sum')
    value_sum: float
    time_sum: float

    def reset(self) -> None:
        self.value_sum = 0.0
        self.time_sum = 0.0

    def get(self) -> float:
        return self.value_sum / self.time_sum

    def aggregate(self, start: datetime.datetime, end: datetime.datetime, value: Union[float, int]) -> None:
        delta_seconds = (end - start).total_seconds()
        self.time_sum += delta_seconds
        self.value_sum += value * delta_seconds


class MinAggregator(AbstractAggregator[Union[float, int], float]):
    """
    Aggregator implementation for AggregationMethod.MINIMUM. It calculates the minimum of all values in an aggregation
    interval. The value's begin..end interval is ignored. Thus, even values with 0 duration are taken into account.
    """
    __slots__ = ('value',)
    value: float

    def reset(self) -> None:
        self.value = float('inf')

    def get(self) -> float:
        return self.value

    def aggregate(self, start: datetime.datetime, end: datetime.datetime, value: Union[int, float]) -> None:
        self.value = min(self.value, value)


class MaxAggregator(AbstractAggregator[Union[float, int], float]):
    """
    Aggregator implementation for AggregationMethod.MAXIMUM. It calculates the maximum of all values in an aggregation
    interval. The value's begin..end interval is ignored. Thus, even values with 0 duration are taken into account.
    """
    __slots__ = ('value',)
    value: float

    def reset(self) -> None:
        self.value = float('-inf')

    def get(self) -> float:
        return self.value

    def aggregate(self, start: datetime.datetime, end: datetime.datetime, value: Union[int, float]) -> None:
        self.value = max(self.value, value)


class OnTimeAggregator(AbstractAggregator[Any, float]):
    """
    Aggregator implementation for AggregationMethod.ONTIME. It calculates the aggregated duration of all intervals where
    the value evaluates to True. The value is returned in seconds.
    """
    __slots__ = ('value',)
    value: datetime.timedelta

    def reset(self) -> None:
        self.value = datetime.timedelta(0)

    def get(self) -> float:
        return self.value.total_seconds()

    def aggregate(self, start: datetime.datetime, end: datetime.datetime, value: bool) -> None:
        if value:
            self.value += end - start


class OnTimeRatioAggregator(AbstractAggregator[Any, float]):
    """
    # Aggregator implementation for AggregationMethod.ONTIME. It calculates the ratio of all intervals where the value
    evaluates to True to the total duration of the aggregation interval.
    """
    __slots__ = ('value', 'total_time')
    value: datetime.timedelta
    total_time: datetime.timedelta

    def reset(self) -> None:
        self.value = datetime.timedelta(0)
        self.total_time = datetime.timedelta(0)

    def get(self) -> float:
        return self.value / self.total_time

    def aggregate(self, start: datetime.datetime, end: datetime.datetime, value: Any) -> None:
        delta = end - start
        self.total_time += delta
        if value:
            self.value += delta



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import logging
import math
from typing import NamedTuple, Union, overload

from shc.conversion import register_converter, get_converter

logger = logging.getLogger(__name__)


class RangeFloat1(float):
    """
    A range / percentage value, represented as a floating point number from 0.0 (0%) to 1.0 (100%).
    """
    def __new__(cls, *args, **kwargs):
        # noinspection PyArgumentList
        res = float.__new__(cls, *args, **kwargs)
        if not 0.0 <= res <= 1.0:
            raise ValueError("{} is out of the allowed range for type {}".format(res, cls.__name__))
        return res

    @overload
    def __mul__(self, other: "RangeFloat1") -> "RangeFloat1": ...
    @overload
    def __mul__(self, other: Union[int, float]) -> float: ...

    def __mul__(self, other):
        if isinstance(other, RangeFloat1):
            return RangeFloat1(super().__mul__(float(other)))
        else:
            return super().__mul__(other)

    @overload
    def __rmul__(self, other: "RangeFloat1") -> "RangeFloat1": ...
    @overload
    def __rmul__(self, other: Union[int, float]) -> float: ...

    def __rmul__(self, other):
        return self.__mul__(other)


class RangeUInt8(int):
    """
    A range / percentage value, represented as an 8bit integer number from 0 (0%) to 255 (100%).
    """
    @classmethod
    def from_float(cls, value: float) -> "RangeUInt8":
        if not isinstance(value, RangeFloat1):
            value = RangeFloat1(value)
        return RangeUInt8(round(value * 255))

    def as_float(self) -> RangeFloat1:
        return RangeFloat1(int(self) / 255)


class RangeInt0To100(int):
    """
    A range / percentage value, represented as an 8bit integer percent number from 0 (0%) to 100 (100%).
    """
    @classmethod
    def from_float(cls, value: float) -> "RangeInt0To100":
        if not isinstance(value, RangeFloat1):
            value = RangeFloat1(value)
        return RangeInt0To100(round(value * 100))

    def as_float(self) -> RangeFloat1:
        return RangeFloat1(int(self) / 100)


register_converter(RangeFloat1, RangeInt0To100, RangeInt0To100.from_float)
register_converter(RangeUInt8, RangeInt0To100, lambda v: RangeInt0To100.from_float(v.as_float()))
register_converter(RangeUInt8, RangeFloat1, lambda v: v.as_float())
register_converter(RangeInt0To100, RangeFloat1, lambda v: v.as_float())
register_converter(RangeInt0To100, RangeUInt8, lambda v: RangeUInt8.from_float(v.as_float()))
register_converter(RangeFloat1, RangeUInt8, RangeUInt8.from_float)

register_converter(float, RangeFloat1, lambda v: RangeFloat1(v))
register_converter(int, RangeUInt8, lambda v: RangeUInt8(v))
register_converter(int, RangeInt0To100, lambda v: RangeInt0To100(v))
register_converter(RangeFloat1, float, lambda v: float(v))
register_converter(RangeUInt8, int, lambda v: int(v))
register_converter(RangeInt0To100, int, lambda v: int(v))

register_converter(RangeUInt8, bool, lambda v: bool(v))
register_converter(RangeFloat1, bool, lambda v: bool(v))
register_converter(RangeInt0To100, bool, lambda v: bool(v))
register_converter(bool, RangeUInt8, lambda v: RangeUInt8(255 if v else 0))
register_converter(bool, RangeFloat1, lambda v: RangeFloat1(1.0 if v else 0.0))
register_converter(bool, RangeInt0To100, lambda v: RangeInt0To100(100 if v else 0))


class AngleUInt8(int):
    """
    An angle, encoded as a 8-bit integer, from 0 (0°) to 255 (360°).
    """
    pass


class RGBUInt8(NamedTuple):
    """
    A 24bit color in RGB colorspace, composed of three :class:`RangeUInt8` values `red`, `green` and `blue`
    """
    red: RangeUInt8
    green: RangeUInt8
    blue: RangeUInt8

    def dimmed(self, other: Union[RangeUInt8, RangeFloat1, RangeInt0To100]) -> "RGBUInt8":
        if not isinstance(other, RangeFloat1):
            other = get_converter(type(other), RangeFloat1)(other)

        return RGBUInt8(RangeUInt8.from_float(self.red.as_float() * other),
                        RangeUInt8.from_float(self.green.as_float() * other),
                        RangeUInt8.from_float(self.blue.as_float() * other))

    @classmethod
    def from_float(cls, value: "RGBFloat1") -> "RGBUInt8":
        return RGBUInt8(RangeUInt8.from_float(value.red),
                        RangeUInt8.from_float(value.green),
                        RangeUInt8.from_float(value.blue))

    def as_float(self) -> "RGBFloat1":
        return RGBFloat1(self.red.as_float(),
                         self.green.as_float(),
                         self.blue.as_float())


class RGBFloat1(NamedTuple):
    """
    A floating point RGB color, composed of three :class:`RangeFloat1` values `red`, `green` and `blue`
    """
    red: RangeFloat1
    green: RangeFloat1
    blue: RangeFloat1

    def dimmed(self, other: Union[RangeUInt8, RangeFloat1, RangeInt0To100]) -> "RGBFloat1":
        if not isinstance(other, RangeFloat1):
            other = get_converter(type(other), RangeFloat1)(other)

        return RGBFloat1(self.red * other,
                         self.green * other,
                         self.blue * other)


register_converter(RGBUInt8, RGBFloat1, lambda v: v.as_float())
register_converter(RGBFloat1, RGBUInt8, RGBUInt8.from_float)


class HSVFloat1(NamedTuple):
    """
    A floating point color in HSV colorspace, composed of three :class:`RangeFloat1` values `hue`, `saturation` and
    `value`.
    """
    hue: RangeFloat1
    saturation: RangeFloat1
    value: RangeFloat1

    def dimmed(self, other: Union[RangeUInt8, RangeFloat1, RangeInt0To100]) -> "HSVFloat1":
        if not isinstance(other, RangeFloat1):
            other = get_converter(type(other), RangeFloat1)(other)
        return HSVFloat1(self.hue, self.saturation, RangeFloat1(self.value * other))

    @classmethod
    def from_rgb(cls, value: RGBFloat1) -> "HSVFloat1":
        # Taken from Wikipedia: https://de.wikipedia.org/wiki/HSV-Farbraum#Umrechnung_RGB_in_HSV/HSL
        r = value.red
        g = value.green
        b = value.blue
        max_v = max(r, g, b)
        min_v = min(r, g, b)
        h = (0 if max_v == min_v
             else 1/6 * (g - b) / (max_v - min_v) if max_v == r
             else 2/6 + 1/6 * (b - r) / (max_v - min_v) if max_v == g
             else 4/6 + 1/6 * (r - g) / (max_v - min_v))
        s = 0 if max_v == 0 else (max_v - min_v) / max_v
        v = max_v
        if h < 0:
            h += 1
        return cls(RangeFloat1(h), RangeFloat1(s), RangeFloat1(v))

    def as_rgb(self) -> "RGBFloat1":
        # Taken from Wikipedia: https://de.wikipedia.org/wiki/HSV-Farbraum#Umrechnung_HSV_in_RGB
        h = self.hue
        s = self.saturation
        v = self.value
        hi = math.floor(h * 6)
        f = h * 6 - hi
        p = v * (1 - s)
        q = v * (1 - s * f)
        t = v * (1 - s * (1 - f))

        r, g, b = ((v, t, p) if hi in (0, 6)
                   else (q, v, p) if hi == 1
                   else (p, v, t) if hi == 2
                   else (p, q, v) if hi == 3
                   else (t, p, v) if hi == 4
                   else (v, p, q))  # if hi == 5
        return RGBFloat1(RangeFloat1(r), RangeFloat1(g), RangeFloat1(b))


register_converter(HSVFloat1, RGBFloat1, lambda v: v.as_rgb())
register_converter(RGBFloat1, HSVFloat1, HSVFloat1.from_rgb)
register_converter(HSVFloat1, RGBUInt8, lambda v: RGBUInt8.from_float(v.as_rgb()))
register_converter(RGBUInt8, HSVFloat1, lambda v: HSVFloat1.from_rgb(v.as_float()))


class RGBWUInt8(NamedTuple):
    """
    4-channel RGBW LED color value, composed of a :class:`RGBUInt8` for `rgb` and a :class:`RangeUInt8` for `white`.
    """
    rgb: RGBUInt8
    white: RangeUInt8

    def dimmed(self, other: Union[RangeUInt8, RangeFloat1, RangeInt0To100]) -> "RGBWUInt8":
        if not isinstance(other, RangeFloat1):
            other = get_converter(type(other), RangeFloat1)(other)

        return RGBWUInt8(self.rgb.dimmed(other),
                         RangeUInt8.from_float(self.white.as_float() * other))


register_converter(RGBWUInt8, RGBUInt8, lambda x: x.rgb)
register_converter(RGBUInt8, RGBWUInt8, lambda x: RGBWUInt8(x, RangeUInt8(0)))


class CCTUInt8(NamedTuple):
    """
    A CCT LED brightness value, composed of two :class:`RangeUInt8` values `cold` and `warm`
    """
    cold: RangeUInt8
    warm: RangeUInt8

    def dimmed(self, other: Union[RangeUInt8, RangeFloat1, RangeInt0To100]) -> "CCTUInt8":
        if not isinstance(other, RangeFloat1):
            other = get_converter(type(other), RangeFloat1)(other)

        return CCTUInt8(RangeUInt8.from_float(self.cold.as_float() * other),
                        RangeUInt8.from_float(self.warm.as_float() * other))


class RGBCCTUInt8(NamedTuple):
    """
    5 channel LED color value, composed of a :class:`RGBUInt8` for `rgb` and a :class:`CCTUInt8` for `white`.
    """
    rgb: RGBUInt8
    white: CCTUInt8

    def dimmed(self, other: Union[RangeUInt8, RangeFloat1, RangeInt0To100]) -> "RGBCCTUInt8":
        if not isinstance(other, RangeFloat1):
            other = get_converter(type(other), RangeFloat1)(other)

        return RGBCCTUInt8(self.rgb.dimmed(other), self.white.dimmed(other))


register_converter(CCTUInt8, RangeUInt8, lambda x: RangeUInt8((x.cold + x.warm)/2))
register_converter(RangeUInt8, CCTUInt8, lambda x: CCTUInt8(x, x))
register_converter(RGBCCTUInt8, RGBUInt8, lambda x: x.rgb)
register_converter(RGBUInt8, RGBCCTUInt8, lambda x: RGBCCTUInt8(x, CCTUInt8(RangeUInt8(0), RangeUInt8(0))))
register_converter(RGBCCTUInt8, CCTUInt8, lambda x: x.white)
register_converter(CCTUInt8, RGBCCTUInt8, lambda x: RGBCCTUInt8(RGBUInt8(RangeUInt8(0), RangeUInt8(0), RangeUInt8(0)),
                                                                x))
register_converter(RGBCCTUInt8, RGBWUInt8, lambda x: RGBWUInt8(x.rgb, RangeUInt8((x.white.cold + x.white.warm)/2)))
register_converter(RGBWUInt8, RGBCCTUInt8, lambda x: RGBCCTUInt8(x.rgb, CCTUInt8(x.white, x.white)))



import asyncio
import logging
import shutil
import subprocess
import time
import unittest
import unittest.mock
from contextlib import suppress
from typing import List

import asyncio_mqtt  # type: ignore
from paho.mqtt.client import MQTTMessage  # type: ignore

import shc.interfaces.mqtt

from .._helper import InterfaceThreadRunner, async_test, ExampleWritable, AsyncMock


@unittest.skipIf(shutil.which("mosquitto") is None, "mosquitto MQTT broker is not available in PATH")
class MQTTClientTest(unittest.TestCase):
    def setUp(self) -> None:
        self.client_runner = InterfaceThreadRunner(shc.interfaces.mqtt.MQTTClientInterface, "localhost", 42883)
        self.client = self.client_runner.interface
        self.broker_process = subprocess.Popen(["mosquitto", "-p", "42883"])
        time.sleep(0.25)

    def tearDown(self) -> None:
        self.client_runner.stop()
        self.broker_process.terminate()
        self.broker_process.wait()

    @staticmethod
    async def _send_retained_test_message() -> None:
        async with asyncio_mqtt.Client("localhost", 42883, client_id="TestClient") as c:
            await c.publish("test/topic", b"42", 0, True)

    @async_test
    async def test_subscribe(self) -> None:
        await self._send_retained_test_message()

        target_raw = ExampleWritable(bytes).connect(self.client.topic_raw('test/topic'))
        target_raw2 = ExampleWritable(bytes).connect(self.client.topic_raw('test/another/topic'))
        target_str = ExampleWritable(str).connect(self.client.topic_string('test/topic', 'test/#'))
        target_int = ExampleWritable(int).connect(self.client.topic_json(int, 'test/topic'))

        self.client_runner.start()

        await asyncio.sleep(0.50)

        # Due to the double-subscription to test/topic, they may be called multiple times
        target_raw._write.assert_called_with(b'42', unittest.mock.ANY)
        target_raw2._write.assert_not_called()
        target_str._write.assert_called_with('42', unittest.mock.ANY)
        target_int._write.assert_called_with(42, unittest.mock.ANY)
        target_raw._write.reset_mock()
        target_str._write.reset_mock()
        target_int._write.reset_mock()

        async with asyncio_mqtt.Client("localhost", 42883, client_id="TestClient") as c:
            await c.publish("test/topic", b"56", 0, False)

        await asyncio.sleep(0.05)

        target_raw._write.assert_called_once_with(b'56', unittest.mock.ANY)
        target_str._write.assert_called_once_with('56', unittest.mock.ANY)
        target_int._write.assert_called_once_with(56, unittest.mock.ANY)
        target_raw._write.reset_mock()
        target_str._write.reset_mock()
        target_int._write.reset_mock()

        async with asyncio_mqtt.Client("localhost", 42883, client_id="TestClient") as c:
            await c.publish("test/something", b"21", 0, False)

        await asyncio.sleep(0.05)

        target_raw._write.assert_not_called()
        target_str._write.assert_called_once_with('21', unittest.mock.ANY)
        target_raw._write.assert_not_called()

    @async_test
    async def test_publish(self) -> None:
        MESSAGES: List[MQTTMessage] = []

        async def _task() -> None:
            async with asyncio_mqtt.Client("localhost", 42883, client_id="TestClient") as c:
                await c.subscribe('#')
                async with c.unfiltered_messages() as messages:
                    async for msg in messages:
                        MESSAGES.append(msg)

        task = asyncio.create_task(_task())
        await asyncio.sleep(0.1)

        try:
            conn_raw = self.client.topic_raw('test/topic', qos=2)
            conn_str = self.client.topic_string('test/another/topic', 'test/another/#', retain=True)
            conn_json = self.client.topic_json(str, 'test/topic', force_mqtt_subscription=True)
            target_str = ExampleWritable(str).connect(conn_str)

            self.client_runner.start()

            await asyncio.wrap_future(asyncio.run_coroutine_threadsafe(conn_raw.write(b'"500"', [self]),
                                                                       loop=self.client_runner.loop))
            # Write might not wait until the message is published
            await asyncio.sleep(0.1)
            self.assertEqual(1, len(MESSAGES))
            self.assertEqual(b'"500"', MESSAGES[-1].payload)
            self.assertEqual('test/topic', MESSAGES[-1].topic)

            await asyncio.wrap_future(asyncio.run_coroutine_threadsafe(conn_str.write('a test with »«', [self]),
                                                                       loop=self.client_runner.loop))
            # Due to the local subscriber, write() should wait until the message has been received
            target_str._write.assert_called_once_with('a test with »«', [self, conn_str])
            await asyncio.sleep(0.01)
            self.assertEqual(2, len(MESSAGES))
            self.assertEqual('a test with »«'.encode('utf-8'), MESSAGES[-1].payload)
            self.assertEqual('test/another/topic', MESSAGES[-1].topic)

            await asyncio.wrap_future(asyncio.run_coroutine_threadsafe(conn_json.write('text', [self]),
                                                                       loop=self.client_runner.loop))
            # Du to the forced subscription, write() should wait until the message has been received
            await asyncio.sleep(0.01)
            self.assertEqual(3, len(MESSAGES))
            self.assertEqual(b'"text"', MESSAGES[-1].payload)
            self.assertEqual('test/topic', MESSAGES[-1].topic)

        finally:
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

    def test_reconnect(self) -> None:
        asyncio.get_event_loop().run_until_complete(self._send_retained_test_message())

        target_raw = ExampleWritable(bytes).connect(self.client.topic_raw('test/topic'))
        self.client_runner.start()
        # We cannot use ClockMock here, since it does not support asyncio.wait()
        time.sleep(0.05)
        target_raw._write.assert_called_once_with(b'42', unittest.mock.ANY)
        target_raw._write.reset_mock()

        with self.assertLogs("shc.interfaces._helper", logging.ERROR) as ctx:
            self.broker_process.terminate()
            self.broker_process.wait()
            time.sleep(0.2)
        self.assertIn("Disconnected", ctx.output[0])
        self.assertIn("MQTTClientInterface", ctx.output[0])

        # Wait for first reconnect attempt
        with self.assertLogs("shc.interfaces._helper", logging.ERROR) as ctx:
            time.sleep(1.1)
        self.assertIn("Error in interface MQTTClientInterface", ctx.output[0])
        self.assertIn("Connection refused", ctx.output[0])

        # Restart server
        self.broker_process = subprocess.Popen(["mosquitto", "-p", "42883"])

        # Wait for second reconnect attempt
        with unittest.mock.patch.object(self.client.client, 'connect', new=AsyncMock()) as connect_mock:
            time.sleep(1)
            connect_mock.assert_not_called()

        asyncio.get_event_loop().run_until_complete(self._send_retained_test_message())
        time.sleep(5)

        target_raw._write.assert_called_once_with(b'42', unittest.mock.ANY)

    def test_initial_reconnect(self) -> None:
        asyncio.get_event_loop().run_until_complete(self._send_retained_test_message())
        self.client.failsafe_start = True
        target_raw = ExampleWritable(bytes).connect(self.client.topic_raw('test/topic'))

        self.broker_process.terminate()
        self.broker_process.wait()

        # We cannot use ClockMock here, since the server seems to be too slow then

        with self.assertLogs("shc.interfaces._helper", logging.ERROR) as ctx:
            self.client_runner.start()
            time.sleep(0.5)
        self.assertIn("Error in interface MQTTClientInterface", ctx.output[0])
        self.assertIn("Connection refused", ctx.output[0])

        # Restart server
        self.broker_process = subprocess.Popen(["mosquitto", "-p", "42883"])
        time.sleep(0.25)
        asyncio.get_event_loop().run_until_complete(self._send_retained_test_message())

        # wait for reconnect attempt
        with unittest.mock.patch.object(self.client.client, 'connect', new=AsyncMock()) as connect_mock:
            time.sleep(0.15)
            connect_mock.assert_not_called()
        time.sleep(0.4)

        target_raw._write.assert_called_once_with(b'42', unittest.mock.ANY)



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import datetime
import enum
import json
from typing import TypeVar, Dict, Tuple, Type, Callable, Any

S = TypeVar('S')
T = TypeVar('T')

_TYPE_CONVERSIONS: Dict[Tuple[Type, Type], Callable[[Any], Any]] = {}
_JSON_CONVERSIONS: Dict[Type, Tuple[Callable[[Any], Any], Callable[[Any], Any]]] = {}


def register_converter(from_type: Type[S], to_type: Type[T], converter: Callable[[S], T]) -> None:
    _TYPE_CONVERSIONS[(from_type, to_type)] = converter


def register_json_conversion(type_: Type[T], to_json: Callable[[T], Any], from_json: Callable[[Any], T]) -> None:
    _JSON_CONVERSIONS[type_] = (to_json, from_json)


def get_converter(from_type: Type[S], to_type: Type[T]) -> Callable[[S], T]:
    """
    Get the default conversion function for converting values from ``from_type`` to ``to_type``.

    :param from_type: The source type
    :param to_type: The target type
    :return: A function, which converts a value of type ``from_type`` to ``to_type``.
    :raises TypeError: If no converter is known for this particular type conversion
    """
    try:
        return _TYPE_CONVERSIONS[(from_type, to_type)]
    except KeyError as e:
        raise TypeError("No converter available to convert {} into {}"
                        .format(from_type.__name__, to_type.__name__)) from e


class SHCJsonEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, enum.Enum):
            return o.value
        if type(o) in _JSON_CONVERSIONS:
            return _JSON_CONVERSIONS[type(o)][0](o)
        return super().default(o)


def from_json(type_: Type[T], value: Any) -> T:
    if issubclass(type_, (bool, int, float, str)):
        return type_(value)  # type: ignore
    if issubclass(type_, enum.Enum):
        return type_(value)  # type: ignore
    if issubclass(type_, tuple) and type_.__annotations__:
        return type_(*(from_json(t, v) for v, (_n, t) in zip(value, type_.__annotations__.items())))  # type: ignore
    try:
        return _JSON_CONVERSIONS[type_][1](value)
    except KeyError as e:
        raise TypeError("No JSON converter available for {}".format(type_.__name__)) from e


register_converter(int, float, lambda v: float(v))
register_converter(float, int, lambda v: round(v))
register_converter(int, str, lambda v: str(v))
register_converter(float, str, lambda v: str(v))
register_converter(str, int, lambda v: int(v))
register_converter(str, float, lambda v: float(v))
register_converter(int, bool, lambda v: bool(v))
register_converter(float, bool, lambda v: bool(v))
register_converter(str, bool, lambda v: bool(v))
register_converter(bytes, str, lambda v: v.decode('utf-8-sig'))
register_converter(str, bytes, lambda v: v.encode('utf-8'))
register_json_conversion(datetime.date, lambda o: o.isoformat(), lambda v: datetime.date.fromisoformat(v))
register_json_conversion(datetime.datetime, lambda o: o.isoformat(), lambda v: datetime.datetime.fromisoformat(v))
register_json_conversion(datetime.timedelta, lambda o: o.total_seconds(), lambda v: datetime.timedelta(seconds=v))



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import datetime
import math
import operator
from typing import Type, Generic, Any, Iterable, Callable, Union, Dict, Tuple

from . import conversion
from .base import Readable, Subscribable, T, Connectable, Writable, S, LogicHandler, UninitializedError
from .datatypes import RangeFloat1, RangeUInt8, HSVFloat1, RGBFloat1, RGBUInt8


class ExpressionBuilder(Connectable[T], metaclass=abc.ABCMeta):
    """
    An abstract base class for any object that may be used in SHC expressions.

    This class defines all the overloaded operators, to create *Connectable* :class:`ExpressionHandler` objects when
    being used in a Python expression. For using this class's features with any *Readable* + *Subscribable* object, use
    the :class:`ExpressionWrapper` subclass. Additionally, each :class:`ExpressionHandler` is also an
    `ExpressionBuilder`, allowing to combine Expressions to even larger Expressions.

    Currently, the following operators and builtin functions are supported:

    * ``+`` (add)
    * ``-`` (sub)
    * ``*`` (mul)
    * ``/`` (truediv)
    * ``//`` (floordiv)
    * ``%`` (mod)
    * ``abs()``
    * ``-()`` (neg)
    * ``ceil()``
    * ``floor()``
    * ``round()``
    * ``==`` (eq)
    * ``!=`` (ne)
    * ``<`` (lt)
    * ``<=`` (le)
    * ``>`` (gt)
    * ``>=`` (ge)

    Additionally, the following methods are provided:

    * :meth:`not`: To be used instead of the ``not`` operator (which cannot be overridden)
    * :meth:`convert`: Creates a "type cast" expression, which converts this object's value to the specified value type,
      using the default converter function from :mod:`shc.conversion`.
    """
    @staticmethod
    def __get_other_type(other: object) -> type:
        if isinstance(other, Readable) and isinstance(other, Subscribable):
            return other.type
        else:
            return type(other)

    def __add__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_ADD:
            return BinaryExpressionHandler(TYPES_ADD[(self.type, other_type)], self, other, operator.add)
        else:
            return NotImplemented

    def __radd__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_ADD:
            return BinaryExpressionHandler(TYPES_ADD[(other_type, self.type)], other, self, operator.add)
        else:
            return NotImplemented

    def __sub__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_SUB:
            return BinaryExpressionHandler(TYPES_SUB[(self.type, other_type)], self, other, operator.sub)
        else:
            return NotImplemented

    def __rsub__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_SUB:
            return BinaryExpressionHandler(TYPES_SUB[(other_type, self.type)], other, self, operator.sub)
        else:
            return NotImplemented

    def __mul__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_MUL:
            return BinaryExpressionHandler(TYPES_MUL[(self.type, other_type)], self, other, operator.mul)
        else:
            return NotImplemented

    def __rmul__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_MUL:
            return BinaryExpressionHandler(TYPES_MUL[(other_type, self.type)], other, self, operator.mul)
        else:
            return NotImplemented

    def __truediv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_TRUEDIV:
            return BinaryExpressionHandler(TYPES_TRUEDIV[(self.type, other_type)], self, other, operator.truediv)
        else:
            return NotImplemented

    def __rtruediv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_TRUEDIV:
            return BinaryExpressionHandler(TYPES_TRUEDIV[(other_type, self.type)], other, self, operator.truediv)
        else:
            return NotImplemented

    def __floordiv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_FLOORDIV:
            return BinaryExpressionHandler(TYPES_FLOORDIV[(self.type, other_type)], self, other, operator.floordiv)
        else:
            return NotImplemented

    def __rfloordiv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_FLOORDIV:
            return BinaryExpressionHandler(TYPES_FLOORDIV[(other_type, self.type)], other, self, operator.floordiv)
        else:
            return NotImplemented

    def __mod__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_MOD:
            return BinaryExpressionHandler(TYPES_MOD[(self.type, other_type)], self, other, operator.floordiv)
        else:
            return NotImplemented

    def __rmod__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_MOD:
            return BinaryExpressionHandler(TYPES_MOD[(other_type, self.type)], other, self, operator.mod)
        else:
            return NotImplemented

    def __abs__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_ABS_NEG:
            return UnaryExpressionHandler(TYPES_ABS_NEG[self.type], self, operator.abs)
        else:
            raise TypeError("{} cannot be used with abs()".format(self))

    def __neg__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_ABS_NEG:
            return UnaryExpressionHandler(TYPES_ABS_NEG[self.type], self, operator.neg)
        else:
            raise TypeError("{} cannot be negated".format(self))

    def __ceil__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_CEIL_FLOOR_ROUND:
            return UnaryExpressionHandler(TYPES_CEIL_FLOOR_ROUND[self.type], self, math.ceil)
        else:
            raise TypeError("{} cannot be rounded".format(self))

    def __floor__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_CEIL_FLOOR_ROUND:
            return UnaryExpressionHandler(TYPES_CEIL_FLOOR_ROUND[self.type], self, math.floor)
        else:
            raise TypeError("{} cannot be rounded".format(self))

    def __round__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_CEIL_FLOOR_ROUND:
            return UnaryExpressionHandler(TYPES_CEIL_FLOOR_ROUND[self.type], self, round)
        else:
            raise TypeError("{} cannot be rounded".format(self))

    def __eq__(self, other) -> "BinaryExpressionHandler":  # type: ignore
        return BinaryExpressionHandler(bool, self, other, operator.eq)

    def __ne__(self, other) -> "BinaryExpressionHandler":  # type: ignore
        return BinaryExpressionHandler(bool, self, other, operator.eq)

    def __lt__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.lt)
        else:
            return NotImplemented

    def __le__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.le)
        else:
            return NotImplemented

    def __gt__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.gt)
        else:
            return NotImplemented

    def __ge__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.ge)
        else:
            return NotImplemented

    def and_(self, other) -> "BinaryCastExpressionHandler":
        return BinaryCastExpressionHandler(bool, self, other, operator.and_)

    def or_(self, other) -> "BinaryCastExpressionHandler":
        return BinaryCastExpressionHandler(bool, self, other, operator.or_)

    def not_(self) -> "UnaryExpressionHandler":
        return UnaryCastExpressionHandler(bool, self, operator.not_)

    def convert(self, type_: Type[T]) -> "UnaryExpressionHandler":
        converter = conversion.get_converter(self.type, type_)
        return UnaryExpressionHandler(type_, self, converter)


class ExpressionWrapper(Readable[T], Subscribable[T], ExpressionBuilder, Generic[T]):
    """
    Wrapper for any *Readable* + *Subscribable* object to equip it with expression-building capabilities, inherited from
    :class:`ExpressionBuilder`
    """
    def __init__(self, wrapped: Subscribable[T]):
        super().__init__()
        self.wrapped = wrapped
        self.type = wrapped.type

    async def read(self) -> T:
        return await self.wrapped.read()  # type: ignore

    def subscribe(self, subscriber: Writable[S], convert: Union[Callable[[T], S], bool] = False):
        return self.wrapped.subscribe(subscriber, convert)

    def trigger(self, target: LogicHandler, synchronous: bool = False) -> LogicHandler:
        return self.wrapped.trigger(target, synchronous=synchronous)


class ExpressionHandler(Readable[T], Subscribable[T], ExpressionBuilder, Generic[T], metaclass=abc.ABCMeta):
    """
    Base class for expression objects, i.e. the result object of an expression built with :class:`ExpressionBuilder`'s
    overloaded operators

    The ExpressionHandler object stores the operator (as a callable) and a reference to each operand. Thereby, it can
    *read* there current values and evaluate the operation/expression at any time with these values. *ExpressionHandler*
    objects are *Readable* (to evaluate the expression's current value on demand) and *Subscribable* (to evaluate and
    publish the expression's new value, when any of the operands is updated).
    """
    def __init__(self, type_: Type[T], operands: Iterable[object]):
        super().__init__()
        self.type = type_
        for operand in operands:
            if isinstance(operand, Subscribable):
                operand.trigger(self.on_change, synchronous=True)

    @abc.abstractmethod
    async def evaluate(self) -> T:
        pass

    async def on_change(self, _value, origin):
        try:
            await self._publish_and_wait(await self.evaluate(), origin)
        except UninitializedError:
            pass

    async def read(self) -> T:
        return await self.evaluate()

    @staticmethod
    def _wrap_static_value(val: Union[S, Readable[S]]) -> Readable[S]:
        class Wrapper(Readable[S]):
            def __init__(self, v: S):
                self.v = v

            async def read(self) -> S:
                return self.v

        if isinstance(val, Readable):
            return val
        else:
            return Wrapper(val)


# ################################# #
# Operator-based ExpressionHandlers #
# ################################# #

class BinaryExpressionHandler(ExpressionHandler[T], Generic[T]):
    def __init__(self, type_: Type[T], a, b, operator_: Callable[[Any, Any], T]):
        self.a = self._wrap_static_value(a)
        self.b = self._wrap_static_value(b)
        self.operator = operator_
        super().__init__(type_, (a, b))

    async def evaluate(self) -> T:
        return self.operator(await self.a.read(), await self.b.read())

    def __repr__(self) -> str:
        return "{}[{}({}, {})]".format(self.__class__.__name__, self.operator.__name__, repr(self.a), repr(self.b))


class BinaryCastExpressionHandler(BinaryExpressionHandler[T]):
    async def evaluate(self) -> T:
        return self.type(await super().evaluate())  # type: ignore


class UnaryExpressionHandler(ExpressionHandler[T], Generic[T]):
    def __init__(self, type_: Type[T], a, operator_: Callable[[Any], T]):
        self.a = self._wrap_static_value(a)
        self.operator = operator_
        super().__init__(type_, (a,))

    async def evaluate(self) -> T:
        return self.operator(await self.a.read())

    def __repr__(self) -> str:
        return "{}[{}({})]".format(self.__class__.__name__, self.operator.__name__, repr(self.a))


class UnaryCastExpressionHandler(UnaryExpressionHandler[T]):
    async def evaluate(self) -> T:
        return self.type(await super().evaluate())  # type: ignore


# ############################# #
# Functional ExpressionHandlers #
# ############################# #

class IfThenElse(ExpressionHandler, Generic[T]):
    """
    A :class:`ExpressionHandler` version of the ``x if condition else y`` python syntax

    This class takes three connectable objects or static values: `condition`, `then` and `otherwise`. `condition` must
    be ``bool`` (or a *Connectable* with value type ``bool``), `then` and `otherwise` must be of the same type (reps.
    have the same value type). If the `condition` evaluates to True, this object evaluates to the value of `then`,
    otherwise it evaluates to the value of `otherwise`.
    """
    def __init__(self, condition: Union[bool, Readable[bool]], then: Union[T, Readable[T]],
                 otherwise: Union[T, Readable[T]]):
        super().__init__(then.type if isinstance(then, Connectable) else type(then), (condition, then, otherwise))
        self.condition: Readable[bool] = self._wrap_static_value(condition)
        self.then = self._wrap_static_value(then)
        self.otherwise = self._wrap_static_value(otherwise)

    async def evaluate(self) -> T:
        return (await self.then.read()) if (await self.condition.read()) else (await self.otherwise.read())


def not_(a) -> Union[bool, UnaryCastExpressionHandler[bool]]:
    """
    Create a :class:`ExpressionHandler` that wraps a *Connectable* or static value and evaluates to ``not x`` for the
    value `x`.

    This is the workaround for Python's limitations on overriding the ``not`` operator.
    """
    if isinstance(a, Readable) and isinstance(a, Subscribable):
        return UnaryCastExpressionHandler(bool, a, operator.not_)
    return not a


def and_(a, b) -> Union[bool, BinaryCastExpressionHandler[bool]]:
    """
    Create a :class:`ExpressionHandler` that wraps two *Connectables* or static values a,b and evaluates to
    ``a and b``.

    This is the workaround for Python's limitations on overriding the ``and`` operator.
    """
    if (not (isinstance(a, Readable) and isinstance(a, Subscribable))
            and not (isinstance(b, Readable) and isinstance(b, Subscribable))):
        return bool(a and b)
    return BinaryCastExpressionHandler(bool, a, b, operator.and_)


def or_(a, b) -> Union[bool, BinaryCastExpressionHandler[bool]]:
    """
    Create a :class:`ExpressionHandler` that wraps two *Connectables* or static values a,b and evaluates to
    ``a or b``.

    This is the workaround for Python's limitations on overriding the ``or`` operator.
    """
    if (not (isinstance(a, Readable) and isinstance(a, Subscribable))
            and not (isinstance(b, Readable) and isinstance(b, Subscribable))):
        return bool(a or b)
    return BinaryCastExpressionHandler(bool, a, b, operator.or_)


# #################### #
# Type Inference Rules #
# #################### #

TYPES_ADD: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (int, float): float,
    (float, int): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (str, str): str,
    (bytes, bytes): bytes,
    (datetime.date, datetime.timedelta): datetime.date,
    (datetime.datetime, datetime.timedelta): datetime.datetime,
    (datetime.timedelta, datetime.date): datetime.date,
    (datetime.timedelta, datetime.datetime): datetime.datetime,
    (datetime.timedelta, datetime.timedelta): datetime.timedelta,
}
TYPES_SUB: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (int, float): float,
    (float, int): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (datetime.date, datetime.date): datetime.timedelta,
    (datetime.datetime, datetime.datetime): datetime.timedelta,
    (datetime.date, datetime.timedelta): datetime.date,
    (datetime.datetime, datetime.timedelta): datetime.datetime,
    (datetime.timedelta, datetime.timedelta): datetime.timedelta,
}
TYPES_MUL: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (int, float): float,
    (float, int): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (str, int): str,
    (int, str): str,
    (bytes, int): bytes,
    (int, bytes): bytes,
    (datetime.timedelta, float): datetime.timedelta,
    (float, datetime.timedelta): datetime.timedelta,
    (datetime.timedelta, int): datetime.timedelta,
    (int, datetime.timedelta): datetime.timedelta,
    (RangeFloat1, float): float,
    (float, RangeFloat1): float,
    (RangeFloat1, int): int,
    (int, RangeFloat1): int,
    (RangeFloat1, RangeFloat1): RangeFloat1
}
TYPES_FLOORDIV: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (float, int): float,
    (int, float): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (datetime.timedelta, datetime.timedelta): int,
    (datetime.timedelta, int): int,
}
TYPES_TRUEDIV: Dict[Tuple[Type, Type], Type] = {
    (int, int): float,
    (float, int): float,
    (int, float): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (datetime.timedelta, datetime.timedelta): float,
    (datetime.timedelta, int): datetime.timedelta,
    (datetime.timedelta, float): datetime.timedelta,
}
TYPES_MOD = {
    (int, int): int,
    (float, int): float,
    (int, float): float,
    (float, float): float,
    (datetime.timedelta, datetime.timedelta): datetime.timedelta,
}
TYPES_ABS_NEG: Dict[Type, Type] = {
    int: int,
    float: float,
    datetime.timedelta: datetime.timedelta,
}
TYPES_CEIL_FLOOR_ROUND: Dict[Type, Type] = {
    int: int,
    float: int,
}
TYPES_LT_LE_GT_GE: Dict[Tuple[Type, Type], Type] = {
    (int, int): bool,
    (int, float): bool,
    (float, int): bool,
    (float, float): bool,
    (datetime.datetime, datetime.datetime): bool,
    (datetime.date, datetime.date): bool,
    (datetime.time, datetime.time): bool,
}



import asyncio
import datetime
import unittest
import unittest.mock

from shc import variables
from shc.expressions import not_, or_, and_
from test._helper import async_test, ExampleWritable


class TestExpressions(unittest.TestCase):
    @async_test
    async def test_expression_1(self):
        var1 = variables.Variable(int, initial_value=5)
        var2 = variables.Variable(int, initial_value=7)
        expression = var1.EX + (-var2.EX)
        expression2 = expression > 5
        subscriber = ExampleWritable(int)
        expression.subscribe(subscriber)
        subscriber2 = ExampleWritable(bool)
        expression2.subscribe(subscriber2)

        self.assertEqual(-2, await expression.read())
        self.assertEqual(False, await expression2.read())

        await var1.write(15, [self])
        await asyncio.sleep(0.01)
        subscriber._write.assert_called_once_with(8, unittest.mock.ANY)
        subscriber2._write.assert_called_once_with(True, unittest.mock.ANY)

    @async_test
    async def test_expression_2(self):
        var1 = variables.Variable(float, initial_value=17.0)
        var2 = variables.Variable(int, initial_value=7)
        expression = (var2.EX * 10) // (var1.EX - 7)
        expression2 = expression > 5
        expression3 = not_(var1.EX < 10)
        expression4 = expression2.and_(expression3)

        self.assertEqual(7, await expression.read())
        self.assertEqual(True, await expression2.read())
        self.assertEqual(True, await expression3.read())
        self.assertEqual(True, await expression4.read())

    @async_test
    async def test_expression_3(self):
        var1 = variables.Variable(bool, initial_value=True)
        var2 = variables.Variable(int, initial_value=42)
        expression = or_(False, var1.EX)
        expression2 = 10 * var2.EX / 2
        expression3 = 320 / (10 - var2.EX)
        expression4 = and_(True, (8 + var2.EX == 50))

        self.assertEqual(True, await expression.read())
        self.assertEqual(210, await expression2.read())
        self.assertEqual(-10.0, await expression3.read())
        self.assertEqual(True, await expression4.read())

    @async_test
    async def test_expression_4(self):
        var1 = variables.Variable(datetime.datetime, initial_value=datetime.datetime(1970, 1, 1))
        var2 = variables.Variable(int, initial_value=-21)
        expression = abs(var2.EX) % 7
        expression2 = datetime.datetime(1970, 1, 31) - var1.EX
        expression3 = 9 // expression

        self.assertEqual(3, await expression.read())
        self.assertEqual(datetime.timedelta(days=30), await expression2.read())
        self.assertEqual(3, await expression3.read())

    def test_type_errors(self):
        var1 = variables.Variable(bool, initial_value=True)
        var2 = variables.Variable(list, initial_value=[42])
        with self.assertRaises(TypeError):
            var1.EX + var2.EX
        with self.assertRaises(TypeError):
            5.5 + var2.EX
        with self.assertRaises(TypeError):
            var1.EX - var2.EX
        with self.assertRaises(TypeError):
            5.5 - var2.EX
        with self.assertRaises(TypeError):
            var2.EX * 5.5
        with self.assertRaises(TypeError):
            5.5 * var2.EX
        with self.assertRaises(TypeError):
            (- var2.EX)
        with self.assertRaises(TypeError):
            abs(var2.EX)
        with self.assertRaises(TypeError):
            var2.EX % 5
        with self.assertRaises(TypeError):
            5 % var2.EX
        with self.assertRaises(TypeError):
            var2.EX > var1.EX
        with self.assertRaises(TypeError):
            var2.EX < var1.EX
        with self.assertRaises(TypeError):
            var2.EX >= 5
        with self.assertRaises(TypeError):
            var2.EX <= 5
        with self.assertRaises(TypeError):
            var2.EX / var1.EX
        with self.assertRaises(TypeError):
            var2.EX // var1.EX
        with self.assertRaises(TypeError):
            5 / var2.EX
        with self.assertRaises(TypeError):
            5 // var2.EX

    @async_test
    async def test_expressions_concurrent_update(self) -> None:
        var1 = variables.Variable(int)
        var2 = variables.Variable(int).connect(var1.EX + 5)
        var1.connect(var2.EX - 5)

        await asyncio.gather(var1.write(42, []), var2.write(56, []))
        self.assertEqual(await var1.read(), await var2.read() - 5)



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import asyncio
import datetime
import logging
import unittest
import unittest.mock
import time
from typing import NamedTuple

import aiohttp

import shc.web
import shc.interfaces.shc_client
from test._helper import ExampleReadable, InterfaceThreadRunner, ExampleWritable, ExampleSubscribable, async_test, \
    ClockMock, AsyncMock


class ExampleType(NamedTuple):
    the_value: float
    is_it_real: bool


class SHCWebsocketClientTest(unittest.TestCase):
    def setUp(self) -> None:
        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080)
        self.client_runner = InterfaceThreadRunner(shc.interfaces.shc_client.SHCWebClient, 'http://localhost:42080')
        self.server = self.server_runner.interface
        self.client = self.client_runner.interface

    def tearDown(self) -> None:
        self.client_runner.stop()
        self.server_runner.stop()

    def test_subscribe(self) -> None:
        self.server.api(int, "foo")
        bar_object = self.server.api(ExampleType, "bar")\
            .connect(ExampleReadable(ExampleType, ExampleType(42, True)))
        bar_source = ExampleSubscribable(ExampleType).connect(bar_object)

        client_foo = self.client.object(int, 'foo')
        client_bar = self.client.object(ExampleType, 'bar')
        foo_target = ExampleWritable(int).connect(client_foo)
        bar_target = ExampleWritable(ExampleType).connect(client_bar)

        self.server_runner.start()
        self.client_runner.start()

        time.sleep(0.05)
        foo_target._write.assert_not_called()
        bar_target._write.assert_called_once_with(ExampleType(42, True), [client_bar])
        self.assertIsInstance(bar_target._write.call_args[0][0], ExampleType)
        bar_target._write.reset_mock()

        asyncio.run_coroutine_threadsafe(bar_source.publish(ExampleType(56, False), [self]),
                                         loop=self.server_runner.loop).result()
        time.sleep(0.05)
        bar_target._write.assert_called_once_with(ExampleType(56, False), [client_bar])

    @async_test
    async def test_subscribe_error(self) -> None:
        self.server.api(int, "foo")
        self.server.api(ExampleType, "bar")

        bar_client = self.client.object(int, 'bar')

        # Creating an option with equal name should return the same object again or raise a type error
        self.assertIs(self.client.object(int, 'bar'), bar_client)
        with self.assertRaises(TypeError):
            self.client.object(str, 'bar')

        # Test raising of connection errors on startup (server is not started yet)
        with self.assertRaises(aiohttp.ClientConnectionError):
            self.client_runner.start()

        self.server_runner.start()

        # Test raising of subscription errors on startup (inexistent api object name)
        # This requires that the object has a local subscriber (otherwise, subscription is skipped)
        another_client = shc.interfaces.shc_client.SHCWebClient('http://localhost:42080')
        foobar = another_client.object(int, 'foobar')
        foobar.subscribe(ExampleWritable(int))
        with self.assertRaises(shc.interfaces.shc_client.WebSocketAPIError):
            await another_client.start()

    def test_read(self) -> None:
        self.server.api(int, "foo")
        self.server.api(ExampleType, "bar")\
            .connect(ExampleReadable(ExampleType, ExampleType(42, True)))

        client_foo = self.client.object(int, 'foo')
        client_bar = self.client.object(ExampleType, 'bar')

        self.server_runner.start()
        self.client_runner.start()

        with self.assertRaises(shc.base.UninitializedError):
            asyncio.run_coroutine_threadsafe(client_foo.read(), loop=self.client_runner.loop).result()

        result = asyncio.run_coroutine_threadsafe(client_bar.read(), loop=self.client_runner.loop).result()
        self.assertIsInstance(result, ExampleType)
        self.assertEqual(ExampleType(42, True), result)

    def test_write(self) -> None:
        server_bar = self.server.api(ExampleType, "bar")
        target = ExampleWritable(ExampleType).connect(server_bar)

        client_bar = self.client.object(ExampleType, 'bar')

        self.server_runner.start()
        self.client_runner.start()

        asyncio.run_coroutine_threadsafe(client_bar.write(ExampleType(42, True), [self]),
                                         loop=self.client_runner.loop).result()
        time.sleep(0.05)
        target._write.assert_called_once_with(ExampleType(42, True), unittest.mock.ANY)
        self.assertIsInstance(target._write.call_args[0][0], ExampleType)

    def test_reconnect(self) -> None:
        self.server.api(ExampleType, "bar")\
            .connect(ExampleReadable(ExampleType, ExampleType(42, True)))

        client_bar = self.client.object(ExampleType, 'bar')
        bar_target = ExampleWritable(ExampleType).connect(client_bar)

        self.server_runner.start()
        self.client_runner.start()

        bar_target._write.assert_called_once_with(ExampleType(42, True), [client_bar])
        bar_target._write.reset_mock()
        # We cannot use ClockMock here, since it does not support asyncio.wait()

        with self.assertLogs("shc.interfaces._helper", logging.ERROR) as ctx:
            self.server_runner.stop()
        self.assertIn("Unexpected shutdown", ctx.output[0])
        self.assertIn("SHCWebClient", ctx.output[0])

        # Re-setup server
        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080)
        self.server = self.server_runner.interface
        self.server.api(ExampleType, "bar")\
            .connect(ExampleReadable(ExampleType, ExampleType(42, True)))

        # Wait for first reconnect attempt
        with self.assertLogs("shc.interfaces._helper", logging.ERROR) as ctx:
            time.sleep(1.1)
        self.assertIn("Error in interface SHCWebClient", ctx.output[0])
        self.assertIn("Cannot connect to host", ctx.output[0])

        # Start server
        self.server_runner.start()

        # Wait for second reconnect attempt
        with unittest.mock.patch.object(self.client._session, 'ws_connect', new=AsyncMock()) as connect_mock:
            time.sleep(1)
            connect_mock.assert_not_called()
        time.sleep(0.3)

        bar_target._write.assert_called_once_with(ExampleType(42, True), [client_bar])

    def test_initial_reconnect(self) -> None:
        self.client.failsafe_start = True
        client_bar = self.client.object(ExampleType, 'bar')
        bar_target = ExampleWritable(ExampleType).connect(client_bar)

        # We cannot use ClockMock here, since the server seems to be too slow then

        with self.assertLogs("shc.interfaces._helper", logging.ERROR) as ctx:
            self.client_runner.start()
            time.sleep(0.5)
        self.assertIn("Error in interface SHCWebClient", ctx.output[0])
        self.assertIn("Cannot connect to host", ctx.output[0])

        # Start server
        self.server_runner.start()

        # Client should still fail due to missing API object
        with self.assertLogs("shc.interfaces._helper", logging.ERROR) as ctx:
            time.sleep(0.6)
        self.assertIn("Error in interface SHCWebClient", ctx.output[0])
        self.assertIn("Failed to subscribe SHC API object 'bar'", ctx.output[0])

        # Re-setup server
        self.server_runner.stop()
        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080)
        self.server = self.server_runner.interface
        self.server.api(ExampleType, "bar") \
            .connect(ExampleReadable(ExampleType, ExampleType(42, True)))
        self.server_runner.start()

        # wait for second reconnect attempt
        with unittest.mock.patch.object(self.client._session, 'ws_connect', new=AsyncMock()) as connect_mock:
            time.sleep(1)
            connect_mock.assert_not_called()
        time.sleep(0.3)

        bar_target._write.assert_called_once_with(ExampleType(42, True), [client_bar])



import asyncio
import datetime
import unittest
from typing import List, Tuple, Optional, Generic

import shc.log.generic
import shc.log.in_memory
from shc.base import T
from ._helper import async_test, ClockMock


time_series_1 = [
    (datetime.datetime(2020, 1, 1, 0, 0, 0), 20.0),
    (datetime.datetime(2020, 1, 1, 0, 0, 5), 40.0),
    (datetime.datetime(2020, 1, 1, 0, 0, 15), 20.0),
    (datetime.datetime(2020, 1, 1, 0, 0, 26), 40.0),
    (datetime.datetime(2020, 1, 1, 0, 0, 27), 40.0),
    (datetime.datetime(2020, 1, 1, 0, 0, 35), 20.0),
]
time_series_2 = [
    (datetime.datetime(2020, 1, 1, 0, 0, 0), False),
    (datetime.datetime(2020, 1, 1, 0, 0, 5), True),
    (datetime.datetime(2020, 1, 1, 0, 0, 15), False),
    (datetime.datetime(2020, 1, 1, 0, 0, 26), True),
    (datetime.datetime(2020, 1, 1, 0, 0, 27), True),
    (datetime.datetime(2020, 1, 1, 0, 0, 35), False),
]


class ExampleLogVariable(shc.log.generic.PersistenceVariable[T], Generic[T]):
    async def _read_from_log(self) -> Optional[T]:
        raise NotImplementedError

    async def _write_to_log(self, value: T) -> None:
        raise NotImplementedError

    def __init__(self, time_series: List[Tuple[datetime.datetime, T]]):
        super().__init__(type(time_series[0][1]), True)
        self.data = time_series

    async def retrieve_log(self, start_time: datetime.datetime, end_time: datetime.datetime,
                           include_previous: bool = True) -> List[Tuple[datetime.datetime, T]]:
        return self.data


class AbstractLoggingTest(unittest.TestCase):
    @async_test
    async def test_maxmin_aggregation(self) -> None:
        variable = ExampleLogVariable(time_series_1)
        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 30),
                                                        aggregation_method=shc.log.AggregationMethod.MAXIMUM,
                                                        aggregation_interval=datetime.timedelta(seconds=10))
        self.assertEqual(2, len(result))
        self.assertAlmostEqual(40.0, result[0][1])
        self.assertAlmostEqual(40.0, result[1][1])
        self.assertAlmostEqual(datetime.datetime(2020, 1, 1, 0, 0, 10), result[0][0])
        self.assertAlmostEqual(datetime.datetime(2020, 1, 1, 0, 0, 20), result[1][0])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 30),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.MINIMUM,
                                                        aggregation_interval=datetime.timedelta(seconds=10))
        self.assertEqual(2, len(result))
        self.assertAlmostEqual(20.0, result[0][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 0),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        aggregation_method=shc.log.AggregationMethod.MINIMUM,
                                                        aggregation_interval=datetime.timedelta(seconds=10))
        self.assertEqual(1, len(result))
        self.assertAlmostEqual(20.0, result[0][1])

    @async_test
    async def test_average_aggregation(self) -> None:
        variable = ExampleLogVariable(time_series_1)
        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 30),
                                                        aggregation_method=shc.log.AggregationMethod.AVERAGE,
                                                        aggregation_interval=datetime.timedelta(seconds=10))
        self.assertEqual(2, len(result))
        self.assertAlmostEqual(30.0, result[0][1])
        self.assertAlmostEqual(28.0, result[1][1])
        self.assertAlmostEqual(datetime.datetime(2020, 1, 1, 0, 0, 10), result[0][0])
        self.assertAlmostEqual(datetime.datetime(2020, 1, 1, 0, 0, 20), result[1][0])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 25),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.AVERAGE,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(10, len(result))
        self.assertAlmostEqual(32.0, result[0][1])  # 25
        self.assertAlmostEqual(40.0, result[1][1])
        self.assertAlmostEqual(40.0, result[2][1])  # 30
        self.assertAlmostEqual(40.0, result[3][1])
        self.assertAlmostEqual(20.0, result[4][1])  # 35
        self.assertAlmostEqual(20.0, result[5][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 0),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        aggregation_method=shc.log.AggregationMethod.AVERAGE,
                                                        aggregation_interval=datetime.timedelta(seconds=10))
        self.assertEqual(1, len(result))
        self.assertAlmostEqual(30.0, result[0][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 20),
                                                        aggregation_method=shc.log.AggregationMethod.AVERAGE,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(4, len(result))
        self.assertAlmostEqual(40.0, result[0][1])
        self.assertAlmostEqual(40.0, result[1][1])
        self.assertAlmostEqual(20.0, result[2][1])
        self.assertAlmostEqual(20.0, result[3][1])

    @async_test
    async def test_ontime_aggregation(self) -> None:
        variable = ExampleLogVariable(time_series_2)
        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 30),
                                                        aggregation_method=shc.log.AggregationMethod.ON_TIME,
                                                        aggregation_interval=datetime.timedelta(seconds=10))
        self.assertEqual(2, len(result))
        self.assertAlmostEqual(5.0, result[0][1])
        self.assertAlmostEqual(4.0, result[1][1])
        self.assertAlmostEqual(datetime.datetime(2020, 1, 1, 0, 0, 10), result[0][0])
        self.assertAlmostEqual(datetime.datetime(2020, 1, 1, 0, 0, 20), result[1][0])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 25),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.ON_TIME,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(10, len(result))
        self.assertAlmostEqual(1.5, result[0][1])
        self.assertAlmostEqual(2.5, result[1][1])
        self.assertAlmostEqual(2.5, result[2][1])
        self.assertAlmostEqual(2.5, result[3][1])
        self.assertAlmostEqual(0.0, result[4][1])
        self.assertAlmostEqual(0.0, result[4][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 0),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        aggregation_method=shc.log.AggregationMethod.ON_TIME_RATIO,
                                                        aggregation_interval=datetime.timedelta(seconds=10))
        self.assertEqual(1, len(result))
        self.assertAlmostEqual(0.5, result[0][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 10),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 20),
                                                        aggregation_method=shc.log.AggregationMethod.ON_TIME_RATIO,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(4, len(result))
        self.assertAlmostEqual(1.0, result[0][1])
        self.assertAlmostEqual(1.0, result[1][1])
        self.assertAlmostEqual(0.0, result[2][1])
        self.assertAlmostEqual(0.0, result[3][1])

    @async_test
    async def test_empty_aggregation(self) -> None:
        variable = ExampleLogVariable(time_series_1)
        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 40),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.MINIMUM,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(4, len(result))
        self.assertEqual(20.0, result[0][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 40),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.AVERAGE,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(4, len(result))
        self.assertEqual(20.0, result[0][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 40),
                                                        end_time=datetime.datetime(2020, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.ON_TIME_RATIO,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(4, len(result))
        self.assertEqual(1.0, result[0][1])

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2019, 1, 1, 0, 0, 40),
                                                        end_time=datetime.datetime(2019, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.MAXIMUM,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(0, len(result))

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2019, 1, 1, 0, 0, 40),
                                                        end_time=datetime.datetime(2019, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.AVERAGE,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(0, len(result))

        result = await variable.retrieve_aggregated_log(start_time=datetime.datetime(2019, 1, 1, 0, 0, 40),
                                                        end_time=datetime.datetime(2019, 1, 1, 0, 0, 50),
                                                        aggregation_method=shc.log.AggregationMethod.ON_TIME,
                                                        aggregation_interval=datetime.timedelta(seconds=2.5))
        self.assertEqual(0, len(result))

    @async_test
    async def test_aggregation_type_error(self) -> None:
        variable = ExampleLogVariable([(datetime.datetime(2020, 1, 1, 0, 0, 0), "foo")])
        with self.assertRaises(TypeError):
            await variable.retrieve_aggregated_log(start_time=datetime.datetime(2020, 1, 1, 0, 0, 40),
                                                   end_time=datetime.datetime(2020, 1, 1, 0, 0, 50),
                                                   aggregation_method=shc.log.AggregationMethod.MINIMUM,
                                                   aggregation_interval=datetime.timedelta(seconds=2.5))


class InMemoryTest(unittest.TestCase):
    @async_test
    async def test_simple(self) -> None:
        var1 = shc.log.in_memory.InMemoryPersistenceVariable(int, datetime.timedelta(seconds=10))
        with ClockMock(datetime.datetime(2020, 1, 1, 0, 0, 0)):
            await var1.write(1, [self])
            await asyncio.sleep(1)
            await var1.write(2, [self])
            await asyncio.sleep(1)
            await var1.write(3, [self])
            await asyncio.sleep(8)
            await var1.write(4, [self])
            await asyncio.sleep(0.5)
            await var1.write(5, [self])

        # Check data retrieval
        data = await var1.retrieve_log(datetime.datetime(2020, 1, 1, 0, 0, 0, microsecond=500000).astimezone(),
                                       datetime.datetime(2020, 11, 20, 0, 0, 0).astimezone())
        self.assertEqual(4, len(data))
        self.assertListEqual([2, 3, 4, 5], [v for _ts, v in data])
        self.assertAlmostEqual(data[0][0], datetime.datetime(2020, 1, 1, 0, 0, 1).astimezone(),
                               delta=datetime.timedelta(milliseconds=50))
        self.assertAlmostEqual(data[-1][0], datetime.datetime(2020, 1, 1, 0, 0, 10, microsecond=500000).astimezone(),
                               delta=datetime.timedelta(milliseconds=50))

        # Check reading
        self.assertEqual(5, await var1.read())

        # Check that old data has been deleted
        self.assertEqual(4, len(var1.data))



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import asyncio
import logging
from typing import Generic, Type, Optional, List, Any, Union

from .base import Writable, T, Readable, Subscribable, UninitializedError, Reading
from .expressions import ExpressionWrapper

logger = logging.getLogger(__name__)

_ALL_VARIABLES: List["Variable"] = []


async def read_initialize_variables() -> None:
    await asyncio.gather(*(variable._init_from_provider() for variable in _ALL_VARIABLES))


class Variable(Writable[T], Readable[T], Subscribable[T], Reading[T], Generic[T]):
    """
    A Variable object for caching and distributing values of a certain type.

    :param type_: The Variable's value type (used for its ``.type`` attribute, i.e. for the *Connectable* type
        checking mechanism)
    :param name: An optional name of the variable. Used for logging and future displaying purposes.
    :param initial_value: An optional initial value for the Variable. If not provided and no default provider is
        set via :meth:`set_provider`, the Variable is initialized with a None value and any :meth:`read` request
        will raise an :exc:`shc.base.UninitializedError` until the first value update is received.
    """
    _stateful_publishing = True

    def __init__(self, type_: Type[T], name: Optional[str] = None, initial_value: Optional[T] = None):
        self.type = type_
        super().__init__()
        self.name = name
        self._value: Optional[T] = initial_value
        self._variable_fields: List["VariableField"] = []

        # Create VariableFields for each typeannotated field of the type if it is typing.NamedTuple-based.
        if issubclass(type_, tuple) and type_.__annotations__:
            for name, field_type in type_.__annotations__.items():
                variable_field = VariableField(self, name, field_type)
                self._variable_fields.append(variable_field)
                setattr(self, name, variable_field)

        _ALL_VARIABLES.append(self)

    async def _write(self, value: T, origin: List[Any]) -> None:
        old_value = self._value
        self._value = value
        if old_value != value:  # if a single field is different, the full value will also be different
            logger.info("New value %s for Variable %s from %s", value, self, origin[:1])
            self._publish(value, origin)
            for field in self._variable_fields:
                field._recursive_publish(getattr(value, field.field),
                                         None if old_value is None else getattr(old_value, field.field),
                                         origin)

    async def read(self) -> T:
        if self._value is None:
            raise UninitializedError("Variable {} is not initialized yet.".format(repr(self)))
        return self._value

    async def _init_from_provider(self) -> None:
        value = await self._from_provider()
        if value is not None:
            assert(self._default_provider is not None)
            await self._write(value, [self._default_provider[0]])

    @property
    def EX(self) -> ExpressionWrapper:
        return ExpressionWrapper(self)

    def __repr__(self) -> str:
        if self.name:
            return "<Variable \"{}\">".format(self.name)
        else:
            return super().__repr__()


class VariableField(Writable[T], Readable[T], Subscribable[T], Generic[T]):
    _stateful_publishing = True

    def __init__(self, parent: Union[Variable, "VariableField"], field: str, type_: Type[T]):
        self.type = type_
        super().__init__()
        self.parent = parent
        self.variable: Variable = parent.variable if hasattr(parent, 'variable') else parent  # type: ignore
        self.field: str = field
        self._variable_fields: List["VariableField"] = []

        # Create VariableFields for each type-annotated field of the type if it is typing.NamedTuple-based.
        if issubclass(type_, tuple) and type_.__annotations__:
            for name, field_type in type_.__annotations__.items():
                variable_field = VariableField(self, name, field_type)
                self._variable_fields.append(variable_field)
                setattr(self, name, variable_field)

    def _recursive_publish(self, new_value: T, old_value: T, origin: List[Any]):
        if old_value != new_value:
            self._publish(new_value, origin)
            for field in self._variable_fields:
                field._recursive_publish(getattr(new_value, field.field),
                                         None if old_value is None else getattr(old_value, field.field),
                                         origin)

    @property
    def _value(self):
        return None if self.parent._value is None else getattr(self.parent._value, self.field)

    async def _write(self, value: T, origin: List[Any]) -> None:
        if self.parent._value is None:
            raise UninitializedError("Cannot set field {} within Variable {}, since it is uninitialized"
                                     .format(self.field, self.variable))
        await self.parent._write(self.parent._value._replace(**{self.field: value}), origin + [self])

    async def read(self) -> T:
        if self.parent._value is None:
            raise UninitializedError("Variable {} is not initialized yet.", repr(self.variable))
        return getattr(self.parent._value, self.field)

    @property
    def EX(self) -> ExpressionWrapper:
        return ExpressionWrapper(self)



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import asyncio
import datetime
import logging
import math
import random
import time
import weakref
from typing import List, Optional, Callable, Any, Type, Union, Tuple, Iterable, Generic

from .base import Subscribable, LogicHandler, Readable, Writable, T, UninitializedError
from .expressions import ExpressionWrapper

logger = logging.getLogger(__name__)

HIGH_YEAR = 2200


class _TimerSupervisor:
    """
    Interface-like class to supervise all timer instances, i.e. start them on startup of SHC and let them gracefully
    shutdown.

    Should be used as a singleton instance.
    """
    def __init__(self):
        self.supervised_timers: List[_AbstractScheduleTimer] = []
        self.timer_tasks: List[asyncio.Task] = []
        self.temporary_tasks: weakref.WeakSet[asyncio.Task] = weakref.WeakSet()

    async def start(self) -> None:
        logger.info("Starting TimerSupervisor with %s timers ...", len(self.supervised_timers))
        self.timer_tasks = list(map(lambda timer: asyncio.create_task(timer.run()), self.supervised_timers))

    async def wait(self) -> None:
        await asyncio.gather(*self.timer_tasks, return_exceptions=True)

    async def stop(self) -> None:
        logger.info("Cancelling supervised timers ...")
        for task in self.timer_tasks:
            task.cancel()
        for task in self.temporary_tasks:
            task.cancel()

    def register_timer(self, timer: "_AbstractScheduleTimer") -> None:
        self.supervised_timers.append(timer)

    def add_temporary_task(self, task: asyncio.Task) -> None:
        self.temporary_tasks.add(task)


timer_supervisor = _TimerSupervisor()


async def _logarithmic_sleep(target: datetime.datetime):
    while True:
        diff = (target - datetime.datetime.now().astimezone()).total_seconds()
        if diff < 0.2:
            if diff > 0:
                await asyncio.sleep(diff)
            return
        else:
            await asyncio.sleep(diff / 2)


def _random_time(maximum_offset: Optional[datetime.timedelta], random_function: str = 'uniform') -> datetime.timedelta:
    """
    Generate a random timedelta within a given range.

    :param maximum_offset: The maximum absolute value of the random timedelta. Depending on the random function, this
        might not be interpreted strictly (e.g. for the 'gauss' function, the value does only hit the interval by 96%
        chance). If None, return value is 0 seconds.
    :param random_function: The random function / random distribution to use. Currently supported are 'uniform' and
        'gauss'.
    :return: A timedelta roughly between -maximum_offset and +maximum_offset.
    """
    if not maximum_offset:
        return datetime.timedelta()
    if random_function == 'uniform':
        random_value = random.uniform(-1, 1)
    elif random_function == 'gauss':
        random_value = random.gauss(0, 0.5)
    else:
        raise ValueError("Unsupported random function '{}'".format(random_function))
    return maximum_offset * random_value


class _AbstractScheduleTimer(Subscribable[None], metaclass=abc.ABCMeta):
    """
    Abstract base class for all schedule Timer objects.

    These objects are started upon startup of SHC by the :class:`_TimerSupervisor` and enter into a loop where they
    calculate the next trigger time according to some schedule (dependent on the specific subclass) and await that
    point in time via an asyncio sleep.

    Since the Timer's main loop works the same for every schedule Timer implementation, it is implemented in this
    abstract class. Only the :meth:`_next_execution` method has to be overriden by each subclass to calculate the next
    trigger time (or return None if the Timer object finished its task).

    :ivar:`last_execution`: Timestamp of the last execution. May be used by the *_next_execution* method.
    """
    type = type(None)

    def __init__(self):
        super().__init__()
        timer_supervisor.register_timer(self)
        self.last_execution: Optional[datetime.datetime] = None

    async def run(self):
        while True:
            next_execution = self._next_execution()
            if next_execution is None:
                logger.info("Timer %s has fulfilled its job and is quitting now.", self)
                return
            logger.info("Scheduling next execution of timer %s for %s", self, next_execution)
            await _logarithmic_sleep(next_execution)
            self.last_execution = next_execution
            self._publish(None, [])  # we use asynchronous publishing, so no worries about timing

    @abc.abstractmethod
    def _next_execution(self) -> Optional[datetime.datetime]:
        pass


class Every(_AbstractScheduleTimer):
    """
    A schedule timer that periodically triggers with a given interval, optionally extended/shortened by a random time.

    It may either be triggered once on startup of SHC and then enter its periodical loop *or* be aligned with the wall
    clock, i.e. be triggered when the current time in the UNIX era is multiple of the interval. This alignment is
    enabled by default. It has the advantage of keeping the length of the interval, even when the SHC application
    is restarted—als long as application's downtime does not fall together with the application's downtime or in the gap
    between calculated time and (random) offset. In this case the trigger execution will be skipped once.

    Thus, if it is important that the trigger being executed with *at least* the given interval, the *align*
    functionality should be turned off. Otherwise it will cause a more stable interval.

    :param delta: The interval between trigger events
    :param align: If True, the interval is kept even after restarts. If False, the timer triggers once upon
        application start up and keeps the interval from that point on.
    :param offset: An offset to add to the trigger times. May be negative. This can be used for explicitly
        offsetting multiple `Every` timers with the same `delta`.
    :param random: A maximum offset to add to or substract from each trigger time. A random timedelta (roughly)
        between -`random` and +`random` will be added to each individual trigger time.
    :param random_function: Identifier of a random distribution function to be used for calculating the `random`
        offset. See :func:`_random_time`'s documentation for supported values.
    """
    def __init__(self, delta: datetime.timedelta, align: bool = True,
                 offset: datetime.timedelta = datetime.timedelta(), random: Optional[datetime.timedelta] = None,
                 random_function: str = 'uniform'):
        super().__init__()
        self.delta = delta
        self.align = align
        self.offset = offset
        self.random = random
        self.random_function = random_function

    def _next_execution(self) -> Optional[datetime.datetime]:
        if self.align:
            delta_seconds = self.delta.total_seconds()
            now_timestamp = datetime.datetime.now().timestamp()
            next_execution = datetime.datetime.fromtimestamp(
                (math.floor(now_timestamp / delta_seconds) + 1) * delta_seconds)\
                .astimezone()
        else:
            if self.last_execution is None:
                next_execution = datetime.datetime.now().astimezone()
            else:
                next_execution = self.last_execution + self.delta
        return next_execution + self.offset + _random_time(self.random, self.random_function)


def every(*args, **kwargs) -> Callable[[LogicHandler], LogicHandler]:
    """
    Decorator for logic handlers to instantiate a :class:`Every` timer and let it trigger the decorated logic handler.

    All positional and keyword arguments are passed to :meth:`Every.__init__`.
    """
    return Every(*args, **kwargs).trigger


class Once(_AbstractScheduleTimer):
    """
    A schedule timer which only triggers only once at startup of the SHC application, optionally delayed by some static
    offset and an additional random offset.

    :param offset: The single trigger event is delayed by this timedelta after SHC application startup
    :param random: A maximum offset to add to or substract from each trigger time. A random timedelta (roughly)
        between -`random` and +`random` will be added to the trigger time. If the resulting trigger time lies in the
        past, the trigger will executed directly upon application startup.
    :param random_function: Identifier of a random distribution function to be used for calculating the `random`
        offset. See :func:`_random_time`'s documentation for supported values.
    """
    def __init__(self, offset: datetime.timedelta = datetime.timedelta(), random: Optional[datetime.timedelta] = None,
                 random_function: str = 'uniform'):
        super().__init__()
        self.offset = offset
        self.random = random
        self.random_function = random_function
        self.is_executed = False

    def _next_execution(self) -> Optional[datetime.datetime]:
        if self.is_executed:
            return None
        self.is_executed = True
        return datetime.datetime.now().astimezone() + self.offset + _random_time(self.random, self.random_function)


def once(*args, **kwargs) -> Callable[[LogicHandler], LogicHandler]:
    """
    Decorator for logic handlers to instantiate a :class:`Once` timer and let it trigger the decorated logic handler.

    All positional and keyword arguments are passed to :meth:`Once.__init__`.
    """
    return Once(*args, **kwargs).trigger


class EveryNth(int):
    """
    A special integer class to be used as an argument for :meth:`At.__init__` to specify that every nth number of the as
    a valid number of the specific field for triggering the timer.

    E.g. ``month=EveryNth(2)`` equals ``month=[1,3,5,7,9,11]``, ``hour=EveryNth(6) equals hour=[0,6,12,18]``
    """
    pass


ValSpec = Union[int, Iterable[int], EveryNth, None]


class At(_AbstractScheduleTimer):
    """
    Periodic timer which triggers on specific datetime values according to spec based on the Gregorian calendar and wall
    clock times. For each field in (year, month, day, hour, minute, second, millisecond) or (year, week, weekday, hour,
    minute, second, millisecond), a pattern may be specified, which may be

    * `None` (all values allowed for that field)
    * a single `int` value
    * a sorted `list` of valid int values
    * an :class:`EveryNth` object (which is a wrapper around integers): It specifies that every nth value is valid
      (e.g. month=EveryNth(2) equals month=[1,3,5,7,9,11], hour=EveryNth(6) equals hour=[0,6,12,18])

    The timer is scheduled for the next datetime matching the specified pattern of each field.

    :param year: A pattern for the `year` value of the next trigger date/time (range: 0–2200)
    :param month: A pattern for the `month` value of the next trigger date/time (range: 1–12). Must not be used
        together with `weeknum` or `weekday`.
    :param day: A pattern for the `day of month` value of the next trigger date/time (range: 1–31). Must not be used
        together with `weeknum` or `weekday`.
    :param weeknum: A pattern for the `week of year` value of the next trigger date/time according to the ISO
        weeknumber (range: 1-53).  Must not be used together with `month` or `day`.
    :param weekday: A pattern for the `weekday` value of the next trigger date/time (range: 1(monday) - 7(sunday)).
        Must not be used together with `month` or `day`.
    :param hour: A pattern for the `hour` value of the next trigger time (range: 0–23).
    :param minute: A pattern for the `minute` value of the next trigger time (range: 0–59).
    :param second: A pattern for the `second` value of the next trigger time (range: 0–59).
    :param millis: A pattern for the `millisecond` value of the next trigger time (range: 0–999).
    :param random: A maximum offset to add to or substract from each trigger time. A random timedelta (roughly)
        between -`random` and +`random` will be added to each individual trigger time.
    :param random_function: Identifier of a random distribution function to be used for calculating the `random`
        offset. See :func:`_random_time`'s documentation for supported values.
    """
    def __init__(self,
                 year: ValSpec = None,
                 month: ValSpec = None,
                 day: ValSpec = None,
                 weeknum: ValSpec = None,
                 weekday: ValSpec = None,
                 hour: ValSpec = 0,
                 minute: ValSpec = 0,
                 second: ValSpec = 0,
                 millis: ValSpec = 0,
                 random: Optional[datetime.timedelta] = None,
                 random_function: str = 'uniform'):
        super().__init__()
        if weekday is None and weeknum is None:
            self.spec = (year, month, day, hour, minute, second, millis)
            self.week_mode = False
        elif month is None and day is None:
            self.spec = (year, weeknum, weekday, hour, minute, second, millis)
            self.week_mode = True
        else:
            raise ValueError("At-Timer cannot constrain month/day and week/weekday at the same time. Either month and "
                             "day or weeknum and weekday must be None.")
        self.random = random
        self.random_function = random_function

    def _next_execution(self) -> Optional[datetime.datetime]:
        now = datetime.datetime.now()
        # This method iteratively approaches the next execution timestamp according to the self.spec. For this purpose,
        # timestamps are represented as tuples/lists of seven elements:
        # In week-mode: (year, week, weekday, hour, minute, second, millisencond),
        # in month-mode: (year, month, day, hour, minute, second, millisecond)
        # The algorithm uses some kind of simple backtracking: Starting with the current time, it checks each tuple
        # entry's value for compliance with the spec, beginning at the first entry (year). If it does not match, the

        # The lowest value of each entry
        origin = (1, 1, 1, 0, 0, 0, 0)

        def limit(i: int, val: List[int]) -> int:
            """
            Calculate the highest possible value for the entry with index `i` of a tuple/list-timestamp. This may depend
            on the year, month and week of the current value `val`.
            """
            if i > 2:
                return (23, 59, 59, 999)[i - 3]
            if i == 0:
                return HIGH_YEAR
            elif i == 1 and self.week_mode:
                return datetime.date(val[0], 12, 28).isocalendar()[1]  # 52 or 53 depending on the year
                # 29.12., 30.12. and 31.12. may be in week 1 of the next year.
            elif i == 1:
                return 12
            elif i == 2 and self.week_mode:
                return 7
            else:
                return (datetime.date(val[0] + (1 if val[1] == 12 else 0), 1 if val[1] == 12 else val[1] + 1, 1)
                        - datetime.timedelta(days=1)).day

        if self.week_mode:
            val = [now.year, now.isocalendar()[1], now.isocalendar()[2], now.hour, now.minute, now.second,
                   round(now.microsecond / 1000)]
        else:
            val = [now.year, now.month, now.day, now.hour, now.minute, now.second, round(now.microsecond / 1000)]

        # Start with the first entry (year)
        i = 0
        must_increase = False
        # Iterate through entries until all 7 are matching the spec
        while i < 7:
            # In case the entry's current value matches the spec and we didn't flag it to be increased, move on to the
            # next entry
            if not must_increase and self._matches(val[i], self.spec[i], origin[i]):
                i += 1
                continue
            must_increase = False
            # Otherwise calculate the next higher value according to the spec
            new_val = self._next(val[i], self.spec[i], origin[i])
            # If the next higher value is valid, reset all following entries to their minimum values (origin) and
            # proceed with checking the next entry.
            if new_val is not None and new_val <= limit(i, val):
                val[i] = new_val
                i += 1
                for j in range(i, 7):
                    val[j] = origin[j]
            # If there is no valid higher entry, step back one entry: Reset this and all
            # following entries to their minimum values (origin) and proceed with the previous entry, while ensuring it
            # will be increased.
            else:
                i -= 1
                # If we hit the year limit and thus cannot step back any further, return None to indicate that there
                # is no next execution time
                if i < 0:
                    logger.warning("Could not find a next execution time for %s.", self)
                    return None
                for j in range(i + 1, 7):
                    val[j] = origin[j]
                must_increase = True

        if self.week_mode:
            # Unfortunately, the fromisocalendar() function has only been added in Python 3.8:
            # val_date = datetime.date.fromisocalendar(val[0], val[1], val[2])
            val_date = datetime.datetime.strptime("{}-W{}-1".format(val[0], val[1]), '%G-W%V-%u') \
                       + datetime.timedelta(days=val[2]-1)
            result = datetime.datetime(val_date.year, val_date.month, val_date.day, val[3], val[4], val[5],
                                       val[6] * 1000).astimezone()
        else:
            result = datetime.datetime(val[0], val[1], val[2], val[3], val[4], val[5], val[6] * 1000).astimezone()
        return result + _random_time(self.random, self.random_function)

    @staticmethod
    def _matches(val: int, spec: ValSpec, origin) -> bool:
        """
        Check if an entry of a tuple/list-represented timestamp matches the spec value for that entry.
        :param val: The value of the tuple-timestamp entry
        :param spec: The spec for that entry
        :param origin: The minimum value of that entry. It is used to align EveryNth specs.
        :return: True if the current value matches the spec
        """
        if spec is None:
            return True
        if isinstance(spec, Iterable):
            return val in spec
        if isinstance(spec, EveryNth):
            return (val - origin) % spec == 0
        return val == spec

    @staticmethod
    def _next(val: int, spec: ValSpec, origin) -> Optional[int]:
        """
        Calculate the next higher valuer of a tuple/list-represented timestamp entry that matches the spec for that
        entry.
        :param val: The current value of the tuple-timestamp entry
        :param spec: The spec for that entry
        :param origin: The minimum value of that entry. It is used to align EveryNth specs.
        :return: The next higher matching value or None if there is none.
        """
        if spec is None:
            return val + 1
        if isinstance(spec, Iterable):
            try:
                return next(v for v in spec if v > val)
            except StopIteration:
                return None
        if isinstance(spec, EveryNth):
            return val + spec - (val - origin) % spec
        return spec if spec > val else None


def at(*args, **kwargs) -> Callable[[LogicHandler], LogicHandler]:
    """
    Decorator for logic handlers to instantiate a :class:`At` timer and let it trigger the decorated logic handler.

    All positional and keyword arguments are passed to :meth:`At.__init__`.
    """
    return At(*args, **kwargs).trigger


class _DelayedBool(Subscribable[bool], Readable[bool], metaclass=abc.ABCMeta):
    """
    Abstract base class for boolean-based delay timers.

    All derived classes work in a similar way: They wrap a *Subscribable* object of boolean type and re-publish its
    updates according to different rules:

    * :class:`TOn`: Changes from False → True are delayed, while True → False changes are published directly and cancel
        any pending False → True change
    * :class:`TOff`: Changes from True → False are delayed, while False → True changes are published directly and cancel
        any pending True → False change
    * :class:`TOnOff`: Same behaviour as combining a TOn with a TOff, i.e. all changes are delayed and may be
        superseded by a contrary change (in contrast to a simple :class:`Delay` object).
    * :class:`TPulse`: Each True value triggers a True pulse of exactly ``delay`` length, but only if no pulse is
        currently active (i.e. the current value is False; a pulse is not re-triggerable).
    """
    type = bool

    def __init__(self, wrapped: Subscribable[bool], delay: datetime.timedelta):
        super().__init__()
        self.delay = delay
        self._value = False
        self._change_task: Optional[asyncio.Task] = None
        wrapped.trigger(self._update, synchronous=True)

    @abc.abstractmethod
    async def _update(self, value: bool, origin: List[Any]): ...

    async def read(self) -> bool:
        return self._value

    async def _set_delayed(self, value: bool, origin: List[Any]):
        # Make sure this task is cancelled on shutdown
        current_task = asyncio.current_task()
        assert(current_task is not None)
        timer_supervisor.add_temporary_task(current_task)
        try:
            await _logarithmic_sleep(datetime.datetime.now().astimezone() + self.delay)
        except asyncio.CancelledError:
            return
        self._value = value
        self._publish(value, origin)
        self._change_task = None

    @property
    def EX(self) -> ExpressionWrapper:
        return ExpressionWrapper(self)


class TOn(_DelayedBool):
    """
    Power-up delay for bool-type *Subscribable* objects

    This object wraps a *Subscribable* bool object and applies a power-up / turn-on delay to its value. I.e., a `False`
    value is re-published immediately, a `True` value is delayed for the configured time period. If a `False` is
    received during the turn-on delay period, the `True` value is superseded and the `TOn`'s value does not change to
    `True` at all.

    The `TOn` object is *Subscribable* and *Readable* for publishing and providing the delayed value. Thus, it can be
    used like an :ref:`SHC Expression <expressions>`. To actually use it in an Expression, you can use the :attr:`EX`
    property, which wraps the `TOn` object in a :class:`shc.expressions.ExpressionWrapper`.

    :param wrapped: The *Subscribable* object to be wrapped for delaying its value
    :param delay: The power-up delay time. E.g. `datetime.timedelta(seconds=5)`
    """
    async def _update(self, value: bool, origin: List[Any]):
        if value and self._change_task is None:
            self._change_task = asyncio.create_task(self._set_delayed(True, origin))
        elif not value:
            if self._change_task is not None:
                self._change_task.cancel()
                self._change_task = None
            if self._value:
                self._value = False
                self._publish(False, origin)


class TOff(_DelayedBool):
    """
    Turn-off delay for bool-type *Subscribable* objects

    This object wraps a *Subscribable* bool object and applies a turn-off delay to its value. I.e., a `True`
    value is re-published immediately, a `False` value is delayed for the configured time period. If a `True` is
    received during the turn-off delay period, the `False` value is superseded and the `TOff`'s value does stay `True`
    continuously.

    The `TOff` object is *Subscribable* and *Readable* for publishing and providing the delayed value. Thus, it can be
    used like an :ref:`SHC Expression <expressions>`. To actually use it in an Expression, you can use the :attr:`EX`
    property, which wraps the `TOff` object in a :class:`shc.expressions.ExpressionWrapper`.

    :param wrapped: The *Subscribable* object to be wrapped for delaying its value
    :param delay: The turn-off delay time. E.g. `datetime.timedelta(seconds=5)`
    """
    async def _update(self, value: bool, origin: List[Any]):
        if not value and self._change_task is None:
            self._change_task = asyncio.create_task(self._set_delayed(False, origin))
        elif value:
            if self._change_task is not None:
                self._change_task.cancel()
                self._change_task = None
            if not self._value:
                self._value = True
                self._publish(True, origin)


class TOnOff(_DelayedBool):
    """
    Resettable boolean-delay for *Subscribable* objects

    This object wraps a *Subscribable* bool object and applies a delay to its value. It behaves like a series of a
    :class:`TOn` and a :class:`TOff` timer with the same delay time: All value updates of the wrapped object will be
    delayed. Albeit, in contrast to a :class:`Delay` timer, a value update can be aborted by sending the
    contrary value during the delay period. Thus, short pulses in either direction (fast False→True→False or
    True→False→True toggling) is suppressed in the `TOnOff`'s value.

    Like all delay timers, the `TOnOff` object is *Subscribable* and *Readable* for publishing and providing the delayed
    value. Thus, it can be used like an :ref:`SHC Expression <expressions>`. To actually use it in an Expression, you
    can use the :attr:`EX` property, which wraps the `TOnOff` object in a :class:`shc.expressions.ExpressionWrapper`.

    :param wrapped: The *Subscribable* object to be wrapped for delaying its value
    :param delay: The delay time. E.g. `datetime.timedelta(seconds=5)`
    """
    async def _update(self, value: bool, origin: List[Any]):
        if value == self._value and self._change_task is None:
            return
        if self._change_task is not None:
            self._change_task.cancel()
        self._change_task = asyncio.create_task(self._set_delayed(value, origin))


class TPulse(_DelayedBool):
    """
    Non-retriggerable pulse generator for bool-typed *Subscribable* objects

    This object wraps a *Subscribable* bool object and creates a fixed length pulse (`True` period) on each rising edge
    (False→True) change of its value. The pulse is not retriggerable, i.e. it is not prolonged when a second raising
    edge occurs during the pulse.

    Like all delay timers, the `TPulse` object is *Subscribable* and *Readable* for publishing and providing the pulse
    signal. Thus, it can be used like an :ref:`SHC Expression <expressions>`. To actually use it in an Expression, you
    can use the :attr:`EX` property, which wraps the `TPulse` object in a :class:`shc.expressions.ExpressionWrapper`.

    :param wrapped: The *Subscribable* object to be wrapped for delaying its value
    :param delay: The pulse length. E.g. `datetime.timedelta(seconds=5)`
    """
    def __init__(self, wrapped: Subscribable[bool], delay: datetime.timedelta):
        super().__init__(wrapped, delay)
        self._prev_value = False

    async def _update(self, value: bool, origin: List[Any]):
        rising_edge = value and not self._prev_value
        self._prev_value = value
        if rising_edge and not self._value:
            self._change_task = asyncio.create_task(self._set_delayed(False, origin))
            self._value = True
            self._publish(True, origin)


class Delay(Subscribable[T], Readable[T], Generic[T]):
    """
    A *Readable* and *Subscribable* object which wraps another Subscribable object and re-publishes its updates after
    the time interval specified by ``delay``. It also *provides* the state/value from ``delay`` time ago via ``read()``
    calls.

    :param wrapped: The *Subscribable* object to be wrapped for delaying its value
    :param delay: The delay time. E.g. `datetime.timedelta(seconds=5)`
    :param initial_value: If given, the value which is returned by :meth:`read` before a value has been received from
        the `wrapped` object (and passed the delay period). If not given, :meth:`read` raises an
        :class:`shc.base.UninitializedError` in this case.
    """
    def __init__(self, wrapped: Subscribable[T], delay: datetime.timedelta, initial_value: Optional[T] = None):
        self.type = wrapped.type
        super().__init__()
        self.delay = delay
        self._value: Optional[T] = initial_value
        wrapped.trigger(self._update, synchronous=True)

    async def _update(self, value: T, origin: List[Any]) -> None:
        asyncio.create_task(self.__set_delayed(value, origin))

    async def __set_delayed(self, value: T, origin: List[Any]):
        # Make sure this task is cancelled on shutdown
        current_task = asyncio.current_task()
        assert(current_task is not None)
        timer_supervisor.add_temporary_task(current_task)
        try:
            await _logarithmic_sleep(datetime.datetime.now() + self.delay)
        except asyncio.CancelledError:
            return
        changed = value != self._value
        logger.debug("Value %s for Delay %s is now active and published", value, self)
        self._value = value
        self._publish(value, origin)

    async def read(self) -> T:
        if self._value is None:
            raise UninitializedError("{} is not initialized yet.", repr(self))
        return self._value

    @property
    def EX(self) -> ExpressionWrapper:
        return ExpressionWrapper(self)


class TimerSwitch(Subscribable[bool], Readable[bool]):
    """
    A helper to program something similar to those old-fashioned digital timer switch adaptors

    A `TimerSwitch` is basically a bool Variable with a fancy constructor to let some timers set the variable to True or
    False. Additionally, a `duration` mode is built in, which allows to specify a duration after which the `TimerSwitch`
    switches off (is set to False) automatically – similar to the :class:`TPulse` timer (but with re-trigger-ability).

    A simple usage example for switching on at 8:00 and off at 13:37, but 9:00 and 15:00 on weekends may look like::

        timer_switch = TimerSwitch(on=[At(weekday=[1,2,3,4,5], hour=8),
                                       At(weekday=[6,7], hour=9),
                                   off=[At(weekday=[1,2,3,4,5], hour=13, minute=37),
                                        At(weekday=[6,7], hour=15)]
        my_ceiling_lights.connect(timer_switch)

    To enable or disable the `TimerSwitch`, you may want to take a look at the :class:`shc.misc.BreakableSubscription`.

    :param on: A list of timers which should trigger a 'switch on' action (i.e. set the `TimerSwitch`'s value to True
        when triggering)
    :param off: A list of timers which should trigger a 'switch off' action (i.e. set the `TimerSwitch`'s value to
        False when triggering). Either this parameter or the `duration` (but not both) must be specified.
    :param duration: Period of time, after which the `TimerSwitch` will be switched off automatically after each
        'switch on' event. Must be used as an alternative to the `off` parameter.
    :param duration_random: An optional (maximum) offset to add to or substract from the `duration` period. If
        `duration` is used, a random timedelta between ``-duration_random`` and ``+duration_random`` will be added to
        each individual 'on' period.
    """
    type = bool

    def __init__(self, on: Iterable[Subscribable], off: Optional[Iterable[Subscribable]] = None,
                 duration: Optional[datetime.timedelta] = None, duration_random: Optional[datetime.timedelta] = None):
        super().__init__()
        if not off and duration is None:
            raise ValueError("Either 'off' timer specs or a 'duration' must be specified")
        if off and duration is not None:
            raise ValueError("'off' timer specs and a 'duration' cannot be specified at the same time")
        if duration_random and duration is None:
            raise ValueError("A 'duration_random' value does not make sense without a duration")

        self.value = False
        self.duration = duration
        self.duration_random = duration_random
        self.delay_task: Optional[asyncio.Task] = None

        for on_timer in on:
            on_timer.trigger(self._on, synchronous=True)
        if off:
            for off_timer in off:
                off_timer.trigger(self._off, synchronous=True)

    async def _on(self, _v, origin) -> None:
        if not self.value:
            self.value = True
            self._publish(True, origin)
        if self.duration is not None:
            if self.delay_task is not None:
                self.delay_task.cancel()
            self.delay_task = asyncio.create_task(self._delayed_off(origin))
            timer_supervisor.add_temporary_task(self.delay_task)

    async def _off(self, _v, origin) -> None:
        if self.value:
            self.value = False
            self._publish(False, origin)

    async def _delayed_off(self, origin) -> None:
        assert(self.duration is not None)
        await _logarithmic_sleep(datetime.datetime.now().astimezone() + self.duration
                                 + _random_time(self.duration_random))
        await self._off(False, origin)

    async def read(self) -> bool:
        return self.value

    @property
    def EX(self) -> ExpressionWrapper:
        return ExpressionWrapper(self)


class RateLimitedSubscription(Subscribable[T], Generic[T]):
    """
    A transparent wrapper for `Subscribable` objects, that delays and drops values to make sure that a given maximum
    rate of new values is not exceeded.

    :param wrapped: The Subscribable object to be wrapped
    :param min_interval: The minimal allowed interval between published values in seconds
    """
    def __init__(self, wrapped: Subscribable[T], min_interval: float):
        self.type = wrapped.type
        super().__init__()
        wrapped.trigger(self._new_value, synchronous=True)
        self.min_interval = min_interval
        self._last_publish = 0.0
        self._delay_task: Optional[asyncio.Task] = None
        self._latest_value: T
        self._latest_origin: List[Any]

    async def _new_value(self, value: T, origin: List[Any]) -> None:
        next_allowed_publish = self._last_publish + self.min_interval
        now = time.time()
        if now >= next_allowed_publish:
            self._last_publish = now
            self._publish(value, origin)
        else:
            self._latest_value = value
            self._latest_origin = origin
            if not self._delay_task:
                self._delay_task = asyncio.create_task(self._delayed_publish(next_allowed_publish - now))
                timer_supervisor.add_temporary_task(self._delay_task)

    async def _delayed_publish(self, delay: float):
        await asyncio.sleep(delay)
        self._last_publish = time.time()
        self._delay_task = None
        self._publish(self._latest_value, self._latest_origin)
